#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
import tempfile

from deluge import maketorrent


def check_torrent(filename):
    # Test loading with libtorrent to make sure it's valid
    from deluge._libtorrent import lt

    lt.torrent_info(filename)

    # Test loading with our internal TorrentInfo class
    from deluge.ui.common import TorrentInfo

    TorrentInfo(filename)


class TestMakeTorrent:
    def test_save_multifile(self):
        # Create a temporary folder for torrent creation
        tmp_path = tempfile.mkdtemp()
        with open(os.path.join(tmp_path, 'file_A'), 'wb') as _file:
            _file.write(b'a' * (312 * 1024))
        with open(os.path.join(tmp_path, 'file_B'), 'wb') as _file:
            _file.write(b'b' * (2354 * 1024))
        with open(os.path.join(tmp_path, 'file_C'), 'wb') as _file:
            _file.write(b'c' * (11 * 1024))

        t = maketorrent.TorrentMetadata()
        t.data_path = tmp_path
        tmp_fd, tmp_file = tempfile.mkstemp('.torrent')
        t.save(tmp_file)

        check_torrent(tmp_file)

        os.remove(os.path.join(tmp_path, 'file_A'))
        os.remove(os.path.join(tmp_path, 'file_B'))
        os.remove(os.path.join(tmp_path, 'file_C'))
        os.rmdir(tmp_path)
        os.close(tmp_fd)
        os.remove(tmp_file)

    def test_save_singlefile(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_data = tmp_dir + '/data'
            with open(tmp_data, 'wb') as _file:
                _file.write(b'a' * (2314 * 1024))
            t = maketorrent.TorrentMetadata()
            t.data_path = tmp_data
            tmp_file = tmp_dir + '/.torrent'
            t.save(tmp_file)

            check_torrent(tmp_file)

    def test_save_multifile_padded(self):
        # Create a temporary folder for torrent creation
        tmp_path = tempfile.mkdtemp()
        with open(os.path.join(tmp_path, 'file_A'), 'wb') as _file:
            _file.write(b'a' * (312 * 1024))
        with open(os.path.join(tmp_path, 'file_B'), 'wb') as _file:
            _file.write(b'b' * (2354 * 1024))
        with open(os.path.join(tmp_path, 'file_C'), 'wb') as _file:
            _file.write(b'c' * (11 * 1024))

        t = maketorrent.TorrentMetadata()
        t.data_path = tmp_path
        t.pad_files = True
        tmp_fd, tmp_file = tempfile.mkstemp('.torrent')
        t.save(tmp_file)

        check_torrent(tmp_file)

        os.remove(os.path.join(tmp_path, 'file_A'))
        os.remove(os.path.join(tmp_path, 'file_B'))
        os.remove(os.path.join(tmp_path, 'file_C'))
        os.rmdir(tmp_path)
        os.close(tmp_fd)
        os.remove(tmp_file)



#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from deluge.plugins.pluginbase import WebPluginBase

from .common import get_resource

log = logging.getLogger(__name__)


class WebUI(WebPluginBase):

    scripts = [get_resource('stats.js')]

    # The enable and disable methods are not scrictly required on the WebUI
    # plugins. They are only here if you need to register images/stylesheets
    # with the webserver.
    def enable(self):
        log.debug('Stats Web plugin enabled!')

    def disable(self):
        log.debug('Stats Web plugin disabled!')



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import sys
from unittest import mock

import pytest


@pytest.mark.gtkui
class TestGTK3Common:
    def setUp(self):
        sys.modules['gi.repository'] = mock.MagicMock()

    def tearDown(self):
        pass

    def test_cmp(self):
        from deluge.ui.gtk3.common import cmp

        assert cmp(None, None) == 0
        assert cmp(1, None) == 1
        assert cmp(0, None) == 1
        assert cmp(None, 7) == -1
        assert cmp(None, 'bar') == -1
        assert cmp('foo', None) == 1
        assert cmp('', None) == 1



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
from collections import deque

import deluge.component as component
from deluge.component import Component
from deluge.decorators import overrides
from deluge.ui.client import client
from deluge.ui.console.modes.basemode import BaseMode, mkwin
from deluge.ui.console.modes.torrentlist import torrentview, torrentviewcolumns
from deluge.ui.console.modes.torrentlist.add_torrents_popup import (
    show_torrent_add_popup,
)
from deluge.ui.console.modes.torrentlist.filtersidebar import FilterSidebar
from deluge.ui.console.modes.torrentlist.queue_mode import QueueMode
from deluge.ui.console.modes.torrentlist.search_mode import SearchMode
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.widgets.popup import MessagePopup, PopupsHandler

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)


# Big help string that gets displayed when the user hits 'h'
HELP_STR = """
This screen shows an overview of the current torrents Deluge is managing. \
The currently selected torrent is indicated with a white background. \
You can change the selected torrent using the up/down arrows or the \
PgUp/PgDown keys. Home and End keys go to the first and last torrent \
respectively.

Operations can be performed on multiple torrents by marking them and \
then hitting Enter. See below for the keys used to mark torrents.

You can scroll a popup window that doesn't fit its content (like \
this one) using the up/down arrows, PgUp/PgDown and Home/End keys.

All popup windows can be closed/canceled by hitting the Esc key \
or the 'q' key (does not work for dialogs like the add torrent dialog)

The actions you can perform and the keys to perform them are as follows:

{!info!}'h'{!normal!} - {|indent_pos:|}Show this help
{!info!}'p'{!normal!} - {|indent_pos:|}Open preferences
{!info!}'l'{!normal!} - {|indent_pos:|}Enter Command Line mode
{!info!}'e'{!normal!} - {|indent_pos:|}Show the event log view ({!info!}'q'{!normal!} to go back to overview)

{!info!}'a'{!normal!} - {|indent_pos:|}Add a torrent
{!info!}Delete{!normal!} - {|indent_pos:|}Delete a torrent

{!info!}'/'{!normal!} - {|indent_pos:|}Search torrent names. \
Searching starts immediately - matching torrents are highlighted in \
green, you can cycle through them with Up/Down arrows and Home/End keys \
You can view torrent details with right arrow, open action popup with \
Enter key and exit search mode with '/' key, left arrow or \
backspace with empty search field

{!info!}'f'{!normal!} - {|indent_pos:|}Show only torrents in a certain state
      (Will open a popup where you can select the state you want to see)
{!info!}'q'{!normal!} - {|indent_pos:|}Enter queue mode

{!info!}'S'{!normal!} - {|indent_pos:|}Show or hide the sidebar

{!info!}Enter{!normal!} - {|indent_pos:|}Show torrent actions popup. Here you can do things like \
pause/resume, remove, recheck and so on. These actions \
apply to all currently marked torrents. The currently \
selected torrent is automatically marked when you press enter.

{!info!}'o'{!normal!} - {|indent_pos:|}Show and set torrent options - this will either apply \
to all selected torrents(but not the highlighted one) or currently \
selected torrent if nothing is selected

{!info!}'Q'{!normal!} - {|indent_pos:|}quit deluge-console
{!info!}'C'{!normal!} - {|indent_pos:|}show connection manager

{!info!}'m'{!normal!} - {|indent_pos:|}Mark a torrent
{!info!}'M'{!normal!} - {|indent_pos:|}Mark all torrents between currently selected torrent and last marked torrent
{!info!}'c'{!normal!} - {|indent_pos:|}Clear selection

{!info!}'v'{!normal!} - {|indent_pos:|}Show a dialog which allows you to choose columns to display
{!info!}'<' / '>'{!normal!} - {|indent_pos:|}Change column by which to sort torrents

{!info!}Right Arrow{!normal!} - {|indent_pos:|}Torrent Detail Mode. This includes more detailed information \
about the currently selected torrent, as well as a view of the \
files in the torrent and the ability to set file priorities.

{!info!}'q'/Esc{!normal!} - {|indent_pos:|}Close a popup (Note that 'q' does not work for dialogs \
where you input something
"""


class TorrentList(BaseMode, PopupsHandler):
    def __init__(self, stdscr, encoding=None):
        BaseMode.__init__(
            self, stdscr, encoding=encoding, do_refresh=False, depend=['SessionProxy']
        )
        PopupsHandler.__init__(self)
        self.messages = deque()
        self.last_mark = -1
        self.go_top = False
        self.minor_mode = None

        self.consoleui = component.get('ConsoleUI')
        self.coreconfig = self.consoleui.coreconfig
        self.config = self.consoleui.config
        self.sidebar = FilterSidebar(self, self.config)
        self.torrentview_panel = mkwin(
            curses.COLOR_GREEN,
            curses.LINES - 1,
            curses.COLS - self.sidebar.width,
            0,
            self.sidebar.width,
        )
        self.torrentview = torrentview.TorrentView(self, self.config)

        util.safe_curs_set(util.Curser.INVISIBLE)
        self.stdscr.notimeout(0)

    def torrentview_columns(self):
        return self.torrentview_panel.getmaxyx()[1]

    def on_config_changed(self):
        self.config.save()
        self.torrentview.on_config_changed()

    def toggle_sidebar(self):
        if self.config['torrentview']['show_sidebar']:
            self.sidebar.show()
            self.sidebar.resize_window(curses.LINES - 2, self.sidebar.width)
            self.torrentview_panel.resize(
                curses.LINES - 1, curses.COLS - self.sidebar.width
            )
            self.torrentview_panel.mvwin(0, self.sidebar.width)
        else:
            self.sidebar.hide()
            self.torrentview_panel.resize(curses.LINES - 1, curses.COLS)
            self.torrentview_panel.mvwin(0, 0)
        self.torrentview.update_columns()
        # After updating the columns widths, clear row cache to recreate them
        self.torrentview.cached_rows.clear()
        self.refresh()

    @overrides(Component)
    def start(self):
        self.torrentview.on_config_changed()
        self.toggle_sidebar()

        if self.config['first_run']:
            self.push_popup(
                MessagePopup(self, 'Welcome to Deluge', HELP_STR, width_req=0.65)
            )
            self.config['first_run'] = False
            self.config.save()

        if client.connected():
            self.torrentview.update(refresh=False)

    @overrides(Component)
    def update(self):
        if self.mode_paused():
            return

        if client.connected():
            self.torrentview.update(refresh=True)

    @overrides(BaseMode)
    def resume(self):
        super().resume()

    @overrides(BaseMode)
    def on_resize(self, rows, cols):
        BaseMode.on_resize(self, rows, cols)

        if self.popup:
            self.popup.handle_resize()

        if not self.consoleui.is_active_mode(self):
            return

        self.toggle_sidebar()

    def show_torrent_details(self, tid):
        mode = self.consoleui.set_mode('TorrentDetail')
        mode.update(tid)

    def set_minor_mode(self, mode):
        self.minor_mode = mode
        self.refresh()

    def _show_visible_columns_popup(self):
        self.push_popup(torrentviewcolumns.TorrentViewColumns(self))

    @overrides(BaseMode)
    def refresh(self, lines=None):
        # Something has requested we scroll to the top of the list
        if self.go_top:
            self.torrentview.cursel = 0
            self.torrentview.curoff = 0
            self.go_top = False

        if not lines:
            if not self.consoleui.is_active_mode(self):
                return
            self.stdscr.erase()

        self.add_string(1, self.torrentview.column_string, scr=self.torrentview_panel)

        # Update the status bars
        statusbar_args = {'scr': self.stdscr, 'bottombar_help': True}
        if self.torrentview.curr_filter is not None:
            statusbar_args[
                'topbar'
            ] = '{}    {{!filterstatus!}}Current filter: {}'.format(
                self.statusbars.topbar,
                self.torrentview.curr_filter,
            )

        if self.minor_mode:
            self.minor_mode.set_statusbar_args(statusbar_args)

        self.draw_statusbars(**statusbar_args)

        self.torrentview.update_torrents(lines)

        if self.minor_mode:
            self.minor_mode.update_cursor()
        else:
            util.safe_curs_set(util.Curser.INVISIBLE)

        if not self.consoleui.is_active_mode(self):
            return

        self.stdscr.noutrefresh()
        self.torrentview_panel.noutrefresh()

        if not self.sidebar.hidden():
            self.sidebar.refresh()

        if self.popup:
            self.popup.refresh()

        curses.doupdate()

    @overrides(BaseMode)
    def read_input(self):
        # Read the character
        affected_lines = None
        c = self.stdscr.getch()

        # Either ESC or ALT+<some key>
        if c == util.KEY_ESC:
            n = self.stdscr.getch()
            if n == -1:  # Means it was the escape key
                pass
            else:  # ALT+<some key>
                c = [c, n]

        if self.popup:
            ret = self.popup.handle_read(c)
            if self.popup and self.popup.closed():
                self.pop_popup()
            self.refresh()
            return ret
        if util.is_printable_chr(c):
            if chr(c) == 'Q':
                component.get('ConsoleUI').quit()
            elif chr(c) == 'C':
                self.consoleui.set_mode('ConnectionManager')
                return
            elif chr(c) == 'q':
                self.torrentview.update_marked(self.torrentview.cursel)
                self.set_minor_mode(
                    QueueMode(self, self.torrentview._selected_torrent_ids())
                )
                return
            elif chr(c) == '/':
                self.set_minor_mode(SearchMode(self))
                return

        if self.sidebar.has_focus() and c not in [curses.KEY_RIGHT]:
            self.sidebar.handle_read(c)
            self.refresh()
            return

        if self.torrentview.numtorrents < 0:
            return
        elif self.minor_mode:
            self.minor_mode.handle_read(c)
            return

        affected_lines = None
        # Hand off to torrentview
        if self.torrentview.handle_read(c) == util.ReadState.CHANGED:
            affected_lines = self.torrentview.get_input_result()

        if c == curses.KEY_LEFT:
            if not self.sidebar.has_focus():
                self.sidebar.set_focused(True)
            self.refresh()
            return
        elif c == curses.KEY_RIGHT:
            if self.sidebar.has_focus():
                self.sidebar.set_focused(False)
                self.refresh()
                return
            # We enter a new mode for the selected torrent here
            tid = self.torrentview.current_torrent_id()
            if tid:
                self.show_torrent_details(tid)
                return

        elif util.is_printable_chr(c):
            if chr(c) == 'a':
                show_torrent_add_popup(self)
            elif chr(c) == 'v':
                self._show_visible_columns_popup()
            elif chr(c) == 'h':
                self.push_popup(MessagePopup(self, 'Help', HELP_STR, width_req=0.65))
            elif chr(c) == 'p':
                mode = self.consoleui.set_mode('Preferences')
                mode.load_config()
                return
            elif chr(c) == 'e':
                self.consoleui.set_mode('EventView')
                return
            elif chr(c) == 'S':
                self.config['torrentview']['show_sidebar'] = (
                    self.config['torrentview']['show_sidebar'] is False
                )
                self.config.save()
                self.toggle_sidebar()
            elif chr(c) == 'l':
                self.consoleui.set_mode('CmdLine', refresh=True)
                return

        self.refresh(affected_lines)



#
# Copyright (C) 2015 Calum Lind <calumlind@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from deluge.pluginmanagerbase import PluginManagerBase


class TestPluginManagerBase:
    def test_get_plugin_info(self):
        pm = PluginManagerBase('core.conf', 'deluge.plugin.core')
        for p in pm.get_available_plugins():
            for key, value in pm.get_plugin_info(p).items():
                assert isinstance(key, str)
                assert isinstance(value, str)

    def test_get_plugin_info_invalid_name(self):
        pm = PluginManagerBase('core.conf', 'deluge.plugin.core')
        for key, value in pm.get_plugin_info('random').items():
            result = 'not available' if key in ('Name', 'Version') else ''
            assert value == result

    def test_parse_pkg_info_metadata_2_1(self):
        pkg_info = """Metadata-Version: 2.1
Name: AutoAdd
Version: 1.8
Summary: Monitors folders for .torrent files.
Home-page: http://dev.deluge-torrent.org/wiki/Plugins/AutoAdd
Author: Chase Sterling, Pedro Algarvio
Author-email: chase.sterling@gmail.com, pedro@algarvio.me
License: GPLv3
Platform: UNKNOWN

Monitors folders for .torrent files.
        """
        plugin_info = PluginManagerBase.parse_pkg_info(pkg_info)
        for value in plugin_info.values():
            assert value != ''
        result = 'Monitors folders for .torrent files.'
        assert plugin_info['Description'] == result



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import pytest

import deluge.common
import deluge.ui.web.auth
import deluge.ui.web.server
from deluge import configmanager
from deluge.conftest import BaseTestCase
from deluge.ui.web.server import DelugeWeb

from .common import ReactorOverride


@pytest.mark.usefixtures('daemon', 'component')
class WebServerTestBase(BaseTestCase):
    """
    Base class for tests that need a running webapi

    """

    def set_up(self):
        self.host_id = None
        deluge.ui.web.server.reactor = ReactorOverride()
        return self.start_webapi(None)

    def start_webapi(self, arg):
        self.webserver_listen_port = 8999

        config_defaults = deluge.ui.web.server.CONFIG_DEFAULTS.copy()
        config_defaults['port'] = self.webserver_listen_port
        self.config = configmanager.ConfigManager('web.conf', config_defaults)

        self.deluge_web = DelugeWeb(daemon=False)

        host = list(self.deluge_web.web_api.hostlist.config['hosts'][0])
        host[2] = self.listen_port
        self.deluge_web.web_api.hostlist.config['hosts'][0] = tuple(host)
        self.host_id = host[0]
        self.deluge_web.start()


class WebServerMockBase:
    """
    Class with utility functions for mocking with tests using the webserver

    """

    def mock_authentication_ignore(self, auth):
        def check_request(request, method=None, level=None):
            pass

        self.patch(auth, 'check_request', check_request)



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import os.path

import pytest
import pytest_twisted

import deluge.component as component
import deluge.ui.tracker_icons
from deluge.conftest import BaseTestCase
from deluge.ui.tracker_icons import TrackerIcon, TrackerIcons

from . import common

common.disable_new_release_check()


@pytest.mark.internet
class TestTrackerIcons(BaseTestCase):
    def set_up(self):
        # Disable resizing with Pillow for consistency.
        self.patch(deluge.ui.tracker_icons, 'Image', None)
        self.icons = TrackerIcons()

    def tear_down(self):
        return component.shutdown()

    @pytest_twisted.ensureDeferred
    async def test_get_deluge_png(self, mock_mkstemp):
        # Deluge has a png favicon link
        icon = TrackerIcon(common.get_test_data_file('deluge.png'))
        result = await self.icons.fetch('deluge-torrent.org')
        assert result == icon
        assert not os.path.isfile(mock_mkstemp[1])

    @pytest_twisted.ensureDeferred
    async def test_get_google_ico(self):
        # Google doesn't have any icon links
        # So instead we'll grab its favicon.ico
        icon = TrackerIcon(common.get_test_data_file('google.ico'))
        result = await self.icons.fetch('www.google.com')
        assert result == icon

    @pytest_twisted.ensureDeferred
    async def test_get_google_ico_hebrew(self):
        """Test that Google.co.il page is read as UTF-8"""
        icon = TrackerIcon(common.get_test_data_file('google.ico'))
        result = await self.icons.fetch('www.google.co.il')
        assert result == icon

    @pytest_twisted.ensureDeferred
    async def test_get_google_ico_with_redirect(self):
        # google.com redirects to www.google.com
        icon = TrackerIcon(common.get_test_data_file('google.ico'))
        result = await self.icons.fetch('google.com')
        assert result == icon

    @pytest.mark.skip(reason='Site removed favicon, new SNI test will be needed')
    @pytest_twisted.ensureDeferred
    async def test_get_seo_svg_with_sni(self):
        # seo using certificates with SNI support only
        icon = TrackerIcon(common.get_test_data_file('seo.svg'))
        result = await self.icons.fetch('www.seo.com')
        assert result == icon

    @pytest_twisted.ensureDeferred
    async def test_get_empty_string_tracker(self):
        result = await self.icons.fetch('')
        assert result is None

    @pytest_twisted.ensureDeferred
    async def test_invalid_host(self, mock_mkstemp):
        """Test that TrackerIcon can handle invalid hostname"""
        result = await self.icons.fetch('deluge.example.com')
        assert not result
        assert not os.path.isfile(mock_mkstemp[1])



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import pytest
import pytest_twisted

import deluge.component as component
from deluge.core.core import Core
from deluge.core.rpcserver import RPCServer
from deluge.tests import common

common.disable_new_release_check()


@pytest.mark.usefixtures('component')
class TestWebUIPlugin:
    @pytest_twisted.async_yield_fixture(autouse=True)
    async def set_up(self, request):
        self = request.instance
        self.rpcserver = RPCServer(listen=False)
        self.core = Core()
        await component.start()

        yield

        def on_shutdown(result):
            del self.rpcserver
            del self.core

        await component.shutdown().addCallback(on_shutdown)

    def test_enable_webui(self):
        if 'WebUi' not in self.core.get_available_plugins():
            pytest.skip('WebUi plugin not available for testing')

        d = self.core.enable_plugin('WebUi')

        def result_cb(result):
            if 'WebUi' not in self.core.get_enabled_plugins():
                self.fail('Failed to enable WebUi plugin')
            assert result

        d.addBoth(result_cb)
        return d



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from unittest.mock import patch

from deluge.ui.web import auth


class MockConfig:
    def __init__(self, config):
        self.config = config

    def __getitem__(self, key):
        return self.config[key]

    def __setitem__(self, key, value):
        self.config[key] = value


class TestWebAuth:
    @patch('deluge.ui.web.auth.JSONComponent.__init__', return_value=None)
    def test_change_password(self, mock_json):
        config = MockConfig(
            {
                'pwd_sha1': '8d8ff487626141d2b91025901d3ab57211180b48',
                'pwd_salt': '7555d757710158655bd1646e207dee21a89e9226',
            }
        )
        webauth = auth.Auth(config)
        assert webauth.change_password('deluge', 'deluge_new')



#
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
from os import environ

from deluge.ui.ui import UI

log = logging.getLogger(__name__)
# Hide pygame community banner
environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'


# Keep this class in __init__.py to avoid the console having to import everything in gtkui.py
class Gtk(UI):

    cmd_description = """GTK-based graphical user interface"""

    def __init__(self, *args, **kwargs):
        super().__init__(
            'gtk', *args, description='Starts the Deluge GTK+ interface', **kwargs
        )

        group = self.parser.add_argument_group(_('GTK Options'))
        group.add_argument(
            'torrents',
            metavar='<torrent>',
            nargs='*',
            default=None,
            help=_(
                'Add one or more torrent files, torrent URLs or magnet URIs'
                ' to a currently running Deluge GTK instance'
            ),
        )

    def start(self):
        super().start()
        import deluge.common

        from .gtkui import GtkUI

        def run(options):
            try:
                gtkui = GtkUI(options)
                gtkui.start()
            except Exception as ex:
                log.exception(ex)
                raise

        deluge.common.run_profiled(
            run,
            self.options,
            output_file=self.options.profile,
            do_profile=self.options.profile,
        )


def start():
    Gtk().start()



#
# Copyright (C) 2013 Bro <bro.development@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.ui.client import client

from .path_combo_chooser import PathChooserComboBox

log = logging.getLogger(__name__)


def singleton(cls):
    instances = {}

    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]

    return getinstance


@singleton
class PathChoosersHandler(component.Component):
    def __init__(self, paths_config_key=None):
        # self.chooser_name = "PathChooser_%d" % (len(PathChooser.path_choosers) +1)
        component.Component.__init__(self, 'PathChoosersHandler')
        self.path_choosers = []
        self.paths_list_keys = []
        self.config_properties = {}
        self.started = False
        self.config_keys_to_funcs_mapping = {
            'path_chooser_show_chooser_button_on_localhost': 'filechooser_button_visible',
            'path_chooser_show_path_entry': 'path_entry_visible',
            'path_chooser_auto_complete_enabled': 'auto_complete_enabled',
            'path_chooser_show_folder_name': 'show_folder_name_on_button',
            'path_chooser_accelerator_string': 'accelerator_string',
            'path_chooser_show_hidden_files': 'show_hidden_files',
            'path_chooser_max_popup_rows': 'max_popup_rows',
        }

    def start(self):
        self.started = True
        self.update_config_from_core()

    def stop(self):
        self.started = False

    def update_config_from_core(self):
        def _on_config_values(config):
            self.config_properties.update(config)
            for chooser in self.path_choosers:
                chooser.set_config(config)

        keys = list(self.config_keys_to_funcs_mapping)
        keys += self.paths_list_keys
        client.core.get_config_values(keys).addCallback(_on_config_values)

    def register_chooser(self, chooser):
        chooser.config_key_funcs = {}
        for key in self.config_keys_to_funcs_mapping:
            chooser.config_key_funcs[key] = [None, None]
            chooser.config_key_funcs[key][0] = getattr(
                chooser, 'get_%s' % self.config_keys_to_funcs_mapping[key]
            )
            chooser.config_key_funcs[key][1] = getattr(
                chooser, 'set_%s' % self.config_keys_to_funcs_mapping[key]
            )

        self.path_choosers.append(chooser)
        if chooser.paths_config_key not in self.paths_list_keys:
            self.paths_list_keys.append(chooser.paths_config_key)
            if self.started:
                self.update_config_from_core()
        else:
            chooser.set_config(self.config_properties)

    def set_value_for_path_choosers(self, value, key):
        for chooser in self.path_choosers:
            chooser.config_key_funcs[key][1](value)

        # Save to core
        if key != 'path_chooser_max_popup_rows':
            client.core.set_config({key: value})
        else:
            # Since the max rows value can be changed fast with a spinbutton, we
            # delay saving to core until the values hasn't been changed in 1 second.
            self.max_rows_value_set = value

            def update(value_):
                # The value hasn't been changed in one second, so save to core
                if self.max_rows_value_set == value_:
                    client.core.set_config({'path_chooser_max_popup_rows': value})

            from twisted.internet import reactor

            reactor.callLater(1, update, value)

    def on_list_values_changed(self, values, key, caller):
        # Save to core
        config = {key: values}
        client.core.set_config(config)
        # Set the values on all path choosers with that key
        for chooser in self.path_choosers:
            # Found chooser with values from 'key'
            if chooser.paths_config_key == key:
                chooser.set_values(values)

    def get_config_keys(self):
        keys = list(self.config_keys_to_funcs_mapping)
        keys += self.paths_list_keys
        return keys


class PathChooser(PathChooserComboBox):
    def __init__(self, paths_config_key=None, parent=None):
        self.paths_config_key = paths_config_key
        super().__init__(parent=parent)
        self.chooser_handler = PathChoosersHandler()
        self.chooser_handler.register_chooser(self)
        self.set_auto_completer_func(self.on_completion)
        self.connect('list-values-changed', self.on_list_values_changed_event)
        self.connect(
            'auto-complete-enabled-toggled', self.on_auto_complete_enabled_toggled
        )
        self.connect('show-filechooser-toggled', self.on_show_filechooser_toggled)
        self.connect(
            'show-folder-name-on-button', self.on_show_folder_on_button_toggled
        )
        self.connect('show-path-entry-toggled', self.on_show_path_entry_toggled)
        self.connect('accelerator-set', self.on_accelerator_set)
        self.connect('max-rows-changed', self.on_max_rows_changed)
        self.connect('show-hidden-files-toggled', self.on_show_hidden_files_toggled)

    def on_auto_complete_enabled_toggled(self, widget, value):
        self.chooser_handler.set_value_for_path_choosers(
            value, 'path_chooser_auto_complete_enabled'
        )

    def on_show_filechooser_toggled(self, widget, value):
        self.chooser_handler.set_value_for_path_choosers(
            value, 'path_chooser_show_chooser_button_on_localhost'
        )

    def on_show_folder_on_button_toggled(self, widget, value):
        self.chooser_handler.set_value_for_path_choosers(
            value, 'path_chooser_show_folder_name'
        )

    def on_show_path_entry_toggled(self, widget, value):
        self.chooser_handler.set_value_for_path_choosers(
            value, 'path_chooser_show_path_entry'
        )

    def on_accelerator_set(self, widget, value):
        self.chooser_handler.set_value_for_path_choosers(
            value, 'path_chooser_accelerator_string'
        )

    def on_show_hidden_files_toggled(self, widget, value):
        self.chooser_handler.set_value_for_path_choosers(
            value, 'path_chooser_show_hidden_files'
        )

    def on_max_rows_changed(self, widget, value):
        self.chooser_handler.set_value_for_path_choosers(
            value, 'path_chooser_max_popup_rows'
        )

    def on_list_values_changed_event(self, widget, values):
        self.chooser_handler.on_list_values_changed(values, self.paths_config_key, self)

    def set_config(self, config):
        self.config = config
        for key in self.config_key_funcs:
            if key in config:
                try:
                    self.config_key_funcs[key][1](config[key])
                except TypeError as ex:
                    log.warning('TypeError: %s', ex)

        # Set the saved paths
        if self.paths_config_key and self.paths_config_key in config:
            self.set_values(config[self.paths_config_key])

    def on_completion(self, args):
        def on_paths_cb(args):
            self.complete(args)

        d = client.core.get_completion_paths(args)
        d.addCallback(on_paths_cb)



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
import tempfile
from email.utils import formatdate

import pytest
import pytest_twisted
from twisted.internet import reactor
from twisted.internet.error import CannotListenError
from twisted.web.error import Error, PageRedirect
from twisted.web.http import NOT_MODIFIED
from twisted.web.resource import EncodingResourceWrapper, Resource
from twisted.web.server import GzipEncoderFactory, Site
from twisted.web.util import redirectTo

from deluge.httpdownloader import download_file
from deluge.log import setup_logger

temp_dir = tempfile.mkdtemp()


def fname(name):
    return os.path.join(temp_dir, name)


class RedirectResource(Resource):
    def render(self, request):
        url = self.get_url().encode('utf8')
        return redirectTo(url, request)


class RenameResource(Resource):
    def render(self, request):
        filename = request.args.get(b'filename', [b'renamed_file'])[0]
        request.setHeader(b'Content-Type', b'text/plain')
        request.setHeader(b'Content-Disposition', b'attachment; filename=' + filename)
        return b'This file should be called ' + filename


class AttachmentResource(Resource):
    def render(self, request):
        content_type = b'text/plain'
        charset = request.getHeader(b'content-charset')
        if charset:
            content_type += b'; charset=' + charset
        request.setHeader(b'Content-Type', content_type)
        request.setHeader(b'Content-Disposition', b'attachment')
        append = request.getHeader(b'content-append') or b''
        content = 'Attachment with no filename set{}'.format(append.decode('utf8'))
        return (
            content.encode(charset.decode('utf8'))
            if charset
            else content.encode('utf8')
        )


class TorrentResource(Resource):
    def render(self, request):
        content_type = b'application/x-bittorrent'
        charset = request.getHeader(b'content-charset')
        if charset:
            content_type += b'; charset=' + charset
        request.setHeader(b'Content-Type', content_type)
        request.setHeader(b'Content-Disposition', b'attachment; filename=test.torrent')
        return 'Binary attachment ignore charset 世丕且\n'.encode()


class CookieResource(Resource):
    def render(self, request):
        request.setHeader(b'Content-Type', b'text/plain')
        if request.getCookie(b'password') is None:
            return b'Password cookie not set!'

        if request.getCookie(b'password') == b'deluge':
            return b'COOKIE MONSTER!'

        return request.getCookie('password')


class GzipResource(Resource):
    def getChild(self, path, request):  # NOQA: N802
        return EncodingResourceWrapper(self, [GzipEncoderFactory()])

    def render(self, request):
        message = request.args.get(b'msg', [b'EFFICIENCY!'])[0]
        request.setHeader(b'Content-Type', b'text/plain')
        return message


class PartialDownloadResource(Resource):
    def __init__(self, *args, **kwargs):
        Resource.__init__(self)
        self.render_count = 0

    def render(self, request):
        # encoding = request.requestHeaders._rawHeaders.get('accept-encoding', None)
        if self.render_count == 0:
            request.setHeader(b'content-length', b'5')
        else:
            request.setHeader(b'content-length', b'3')

        # if encoding == "deflate, gzip, x-gzip":
        request.write('abc')
        self.render_count += 1
        return ''


class TopLevelResource(Resource):
    def __init__(self):
        Resource.__init__(self)
        self.putChild(b'cookie', CookieResource())
        self.putChild(b'gzip', GzipResource())
        self.redirect_rsrc = RedirectResource()
        self.putChild(b'redirect', self.redirect_rsrc)
        self.putChild(b'rename', RenameResource())
        self.putChild(b'attachment', AttachmentResource())
        self.putChild(b'torrent', TorrentResource())
        self.putChild(b'partial', PartialDownloadResource())

    def getChild(self, path, request):  # NOQA: N802
        if not path:
            return self
        else:
            return Resource.getChild(self, path, request)

    def render(self, request):
        if request.getHeader(b'If-Modified-Since'):
            request.setResponseCode(NOT_MODIFIED)
        return b'<h1>Deluge HTTP Downloader tests webserver here</h1>'


class TestDownloadFile:
    def get_url(self, path=''):
        return 'http://localhost:%d/%s' % (self.listen_port, path)

    @pytest_twisted.async_yield_fixture(autouse=True)
    async def setUp(self, request):  # NOQA
        self = request.instance
        setup_logger('warning', fname('log_file'))
        self.website = Site(TopLevelResource())
        self.listen_port = 51242
        self.website.resource.redirect_rsrc.get_url = self.get_url
        for dummy in range(10):
            try:
                self.webserver = reactor.listenTCP(self.listen_port, self.website)
            except CannotListenError as ex:
                error = ex
                self.listen_port += 1
            else:
                break
        else:
            raise error

        yield

        await self.webserver.stopListening()

    def assert_contains(self, filename, contents):
        with open(filename, encoding='utf8') as _file:
            try:
                assert _file.read() == contents
            except Exception as ex:
                pytest.fail(ex)
        return filename

    def assert_not_contains(self, filename, contents, file_mode=''):
        with open(filename, encoding='utf8') as _file:
            try:
                assert _file.read() != contents
            except Exception as ex:
                pytest.fail(ex)
        return filename

    @pytest_twisted.ensureDeferred
    async def test_download(self):
        filename = await download_file(self.get_url(), fname('index.html'))
        assert filename == fname('index.html')

    @pytest_twisted.ensureDeferred
    async def test_download_without_required_cookies(self):
        url = self.get_url('cookie')
        filename = await download_file(url, fname('none'))
        self.assert_contains(filename, 'Password cookie not set!')

    @pytest_twisted.ensureDeferred
    async def test_download_with_required_cookies(self):
        url = self.get_url('cookie')
        cookie = {'cookie': 'password=deluge'}
        filename = await download_file(url, fname('monster'), headers=cookie)
        assert filename == fname('monster')
        self.assert_contains(filename, 'COOKIE MONSTER!')

    @pytest_twisted.ensureDeferred
    async def test_download_with_rename(self):
        url = self.get_url('rename?filename=renamed')
        filename = await download_file(url, fname('original'))
        assert filename == fname('renamed')
        self.assert_contains(filename, 'This file should be called renamed')

    @pytest_twisted.ensureDeferred
    async def test_download_with_rename_exists(self):
        open(fname('renamed'), 'w').close()
        url = self.get_url('rename?filename=renamed')
        filename = await download_file(url, fname('original'))
        assert filename == fname('renamed-1')
        self.assert_contains(filename, 'This file should be called renamed')

    @pytest_twisted.ensureDeferred
    async def test_download_with_rename_sanitised(self):
        url = self.get_url('rename?filename=/etc/passwd')
        filename = await download_file(url, fname('original'))
        assert filename == fname('passwd')
        self.assert_contains(filename, 'This file should be called /etc/passwd')

    @pytest_twisted.ensureDeferred
    async def test_download_with_attachment_no_filename(self):
        url = self.get_url('attachment')
        filename = await download_file(url, fname('original'))
        assert filename == fname('original')
        self.assert_contains(filename, 'Attachment with no filename set')

    @pytest_twisted.ensureDeferred
    async def test_download_with_rename_prevented(self):
        url = self.get_url('rename?filename=spam')
        filename = await download_file(url, fname('forced'), force_filename=True)
        assert filename == fname('forced')
        self.assert_contains(filename, 'This file should be called spam')

    @pytest_twisted.ensureDeferred
    async def test_download_with_gzip_encoding(self):
        url = self.get_url('gzip?msg=success')
        filename = await download_file(url, fname('gzip_encoded'))
        self.assert_contains(filename, 'success')

    @pytest_twisted.ensureDeferred
    async def test_download_with_gzip_encoding_disabled(self):
        url = self.get_url('gzip?msg=unzip')
        filename = await download_file(
            url, fname('gzip_encoded'), allow_compression=False
        )
        self.assert_contains(filename, 'unzip')

    @pytest_twisted.ensureDeferred
    async def test_page_redirect_unhandled(self):
        url = self.get_url('redirect')
        with pytest.raises(PageRedirect):
            await download_file(url, fname('none'), handle_redirects=False)

    @pytest_twisted.ensureDeferred
    async def test_page_redirect(self):
        url = self.get_url('redirect')
        filename = await download_file(url, fname('none'), handle_redirects=True)
        assert filename == fname('none')

    @pytest_twisted.ensureDeferred
    async def test_page_not_found(self):
        with pytest.raises(Error):
            await download_file(self.get_url('page/not/found'), fname('none'))

    @pytest.mark.xfail(reason="Doesn't seem like httpdownloader ever implemented this.")
    @pytest_twisted.ensureDeferred
    async def test_page_not_modified(self):
        headers = {'If-Modified-Since': formatdate(usegmt=True)}
        with pytest.raises(Error) as exc_info:
            await download_file(self.get_url(), fname('index.html'), headers=headers)
        assert exc_info.value.status == NOT_MODIFIED

    @pytest_twisted.ensureDeferred
    async def test_download_text_reencode_charset(self):
        """Re-encode as UTF-8 specified charset for text content-type header"""
        url = self.get_url('attachment')
        filepath = fname('test.txt')
        headers = {'content-charset': 'Windows-1251', 'content-append': 'бвгде'}
        filename = await download_file(url, filepath, headers=headers)
        assert filename == filepath
        self.assert_contains(filename, 'Attachment with no filename setбвгде')

    @pytest_twisted.ensureDeferred
    async def test_download_binary_ignore_charset(self):
        """Ignore charset for binary content-type header e.g. torrent files"""
        url = self.get_url('torrent')
        headers = {'content-charset': 'Windows-1251'}
        filepath = fname('test.torrent')
        filename = await download_file(url, fname('test.torrent'), headers=headers)
        assert filename == filepath
        self.assert_contains(filename, 'Binary attachment ignore charset 世丕且\n')



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import pytest_twisted
from twisted.internet.defer import maybeDeferred, succeed
from twisted.internet.task import Clock

import deluge.component as component
import deluge.ui.sessionproxy
from deluge.conftest import BaseTestCase


class Core:
    def __init__(self):
        self.reset()

    def reset(self):
        self.torrents = {}
        self.torrents['a'] = {'key1': 1, 'key2': 2, 'key3': 3}
        self.torrents['b'] = {'key1': 1, 'key2': 2, 'key3': 3}
        self.torrents['c'] = {'key1': 1, 'key2': 2, 'key3': 3}
        self.prev_status = {}

    def get_session_state(self):
        return maybeDeferred(self.torrents.keys)

    def get_torrent_status(self, torrent_id, keys, diff=False):
        if not keys:
            keys = list(self.torrents[torrent_id])

        if not diff:
            ret = {}
            for key in keys:
                ret[key] = self.torrents[torrent_id][key]

            return succeed(ret)

        else:
            ret = {}
            if torrent_id in self.prev_status:
                for key in keys:
                    if (
                        self.prev_status[torrent_id][key]
                        != self.torrents[torrent_id][key]
                    ):
                        ret[key] = self.torrents[torrent_id][key]
            else:
                ret = self.torrents[torrent_id]
            self.prev_status[torrent_id] = dict(self.torrents[torrent_id])
            return succeed(ret)

    def get_torrents_status(self, filter_dict, keys, diff=False):
        if not filter_dict:
            filter_dict['id'] = list(self.torrents)
        if not keys:
            keys = list(self.torrents['a'])
        if not diff:
            if 'id' in filter_dict:
                torrents = filter_dict['id']
                ret = {}
                for torrent in torrents:
                    ret[torrent] = {}
                    for key in keys:
                        ret[torrent][key] = self.torrents[torrent][key]
                return succeed(ret)
        else:
            if 'id' in filter_dict:
                torrents = filter_dict['id']
                ret = {}
                for torrent in torrents:
                    ret[torrent] = {}
                    if torrent in self.prev_status:
                        for key in self.prev_status[torrent]:
                            if (
                                self.prev_status[torrent][key]
                                != self.torrents[torrent][key]
                            ):
                                ret[torrent][key] = self.torrents[torrent][key]
                    else:
                        ret[torrent] = dict(self.torrents[torrent])

                    self.prev_status[torrent] = dict(self.torrents[torrent])
                return succeed(ret)


class Client:
    def __init__(self):
        self.core = Core()

    def __noop__(self, *args, **kwargs):
        return None

    def __getattr__(self, *args, **kwargs):
        return self.__noop__


client = Client()


class TestSessionProxy(BaseTestCase):
    def set_up(self):
        self.clock = Clock()
        self.patch(deluge.ui.sessionproxy, 'time', self.clock.seconds)
        self.patch(deluge.ui.sessionproxy, 'client', client)
        self.sp = deluge.ui.sessionproxy.SessionProxy()
        client.core.reset()
        d = self.sp.start()

        def do_get_torrents_status(torrent_ids):
            inital_keys = ['key1']
            # Advance clock to expire the cache times
            self.clock.advance(2)
            return self.sp.get_torrents_status({'id': torrent_ids}, inital_keys)

        d.addCallback(do_get_torrents_status)
        return d

    def tear_down(self):
        return component.deregister(self.sp)

    def test_startup(self):
        assert client.core.torrents['a'] == self.sp.torrents['a'][1]

    @pytest_twisted.ensureDeferred
    async def test_get_torrent_status_no_change(self):
        result = await self.sp.get_torrent_status('a', [])
        assert result == client.core.torrents['a']

    @pytest_twisted.ensureDeferred
    async def test_get_torrent_status_change_with_cache(self):
        client.core.torrents['a']['key1'] = 2
        result = await self.sp.get_torrent_status('a', ['key1'])
        assert result == {'key1': 1}

    @pytest_twisted.ensureDeferred
    async def test_get_torrent_status_change_without_cache(self):
        client.core.torrents['a']['key1'] = 2
        self.clock.advance(self.sp.cache_time + 0.1)
        result = await self.sp.get_torrent_status('a', [])
        assert result == client.core.torrents['a']

    @pytest_twisted.ensureDeferred
    async def test_get_torrent_status_key_not_updated(self):
        self.clock.advance(self.sp.cache_time + 0.1)
        self.sp.get_torrent_status('a', ['key1'])
        client.core.torrents['a']['key2'] = 99
        result = await self.sp.get_torrent_status('a', ['key2'])
        assert result == {'key2': 99}

    @pytest_twisted.ensureDeferred
    async def test_get_torrents_status_key_not_updated(self):
        self.clock.advance(self.sp.cache_time + 0.1)
        self.sp.get_torrents_status({'id': ['a']}, ['key1'])
        client.core.torrents['a']['key2'] = 99
        result = await self.sp.get_torrents_status({'id': ['a']}, ['key2'])
        assert result == {'a': {'key2': 99}}



#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#               2007-2009 Andrew Resch <andrewresch@gmail.com>
#               2009 Damien Churchill <damoxc@gmail.com>
#               2010 Pedro Algarvio <pedro@algarvio.me>
#               2017 Calum Lind <calumlind+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os.path
from functools import wraps
from sys import exc_info

from pkg_resources import resource_filename


def get_resource(filename):
    return resource_filename(__package__, os.path.join('data', filename))


def raises_errors_as(error):
    """Factory class that returns a decorator which wraps the decorated
    function to raise all exceptions as the specified error type.

    """

    def decorator(func):
        """Returns a function which wraps the given func to raise all exceptions as error."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            """Wraps the function in a try..except block and calls it with the specified args.

            Raises:
                Any exceptions as error preserving the message and traceback.

            """
            try:
                return func(self, *args, **kwargs)
            except Exception:
                (value, tb) = exc_info()[1:]
                raise error(value).with_traceback(tb) from None

        return wrapper

    return decorator


def remove_zeros(ip):
    """Removes unneeded zeros from ip addresses.

    Args:
        ip (str): The ip address.

    Returns:
        str: The ip address without the unneeded zeros.

    Example:
        000.000.000.003 -> 0.0.0.3

    """
    return '.'.join([part.lstrip('0').zfill(1) for part in ip.split('.')])


class BadIP(Exception):
    _message = None

    def __init__(self, message):
        super().__init__(message)

    def __set_message(self, message):
        self._message = message

    def __get_message(self):
        return self._message

    message = property(__get_message, __set_message)
    del __get_message, __set_message


class IP:
    __slots__ = ('q1', 'q2', 'q3', 'q4', '_long')

    def __init__(self, q1, q2, q3, q4):
        self.q1 = q1
        self.q2 = q2
        self.q3 = q3
        self.q4 = q4
        self._long = 0
        for q in self.quadrants():
            self._long = (self._long << 8) | int(q)

    @property
    def address(self):
        return '.'.join([str(q) for q in [self.q1, self.q2, self.q3, self.q4]])

    @property
    def long(self):
        return self._long

    @classmethod
    def parse(cls, ip):
        try:
            q1, q2, q3, q4 = (int(q) for q in ip.split('.'))
        except ValueError:
            raise BadIP(_('The IP address "%s" is badly formed' % ip))
        if q1 < 0 or q2 < 0 or q3 < 0 or q4 < 0:
            raise BadIP(_('The IP address "%s" is badly formed' % ip))
        elif q1 > 255 or q2 > 255 or q3 > 255 or q4 > 255:
            raise BadIP(_('The IP address "%s" is badly formed' % ip))
        return cls(q1, q2, q3, q4)

    def quadrants(self):
        return (self.q1, self.q2, self.q3, self.q4)

    #    def next_ip(self):
    #        (q1, q2, q3, q4) = self.quadrants()
    #        if q4 >= 255:
    #            if q3 >= 255:
    #                if q2 >= 255:
    #                    if q1 >= 255:
    #                        raise BadIP(_('There is not a next IP address'))
    #                    q1 += 1
    #                else:
    #                    q2 += 1
    #            else:
    #                q3 += 1
    #        else:
    #            q4 += 1
    #        return IP(q1, q2, q3, q4)
    #
    #    def previous_ip(self):
    #        (q1, q2, q3, q4) = self.quadrants()
    #        if q4 <= 1:
    #            if q3 <= 1:
    #                if q2 <= 1:
    #                    if q1 <= 1:
    #                        raise BadIP(_('There is not a previous IP address'))
    #                    q1 -= 1
    #                else:
    #                    q2 -= 1
    #            else:
    #                q3 -= 1
    #        else:
    #            q4 -= 1
    #        return IP(q1, q2, q3, q4)

    def __lt__(self, other):
        if isinstance(other, ''.__class__):
            other = IP.parse(other)
        return self.long < other.long

    def __gt__(self, other):
        if isinstance(other, ''.__class__):
            other = IP.parse(other)
        return self.long > other.long

    def __eq__(self, other):
        if isinstance(other, ''.__class__):
            other = IP.parse(other)
        return self.long == other.long

    def __repr__(self):
        return '<{} long={} address="{}">'.format(
            self.__class__.__name__,
            self.long,
            self.address,
        )



#
# Copyright (C) 2009-2010 John Garland <johnnybg+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import re

from deluge.common import decode_bytes

from .common import IP, BadIP, raises_errors_as

log = logging.getLogger(__name__)


class ReaderParseError(Exception):
    pass


class BaseReader:
    """Base reader for blocklist files"""

    def __init__(self, _file):
        """Creates a new BaseReader given a file"""
        self.file = _file

    def open(self):
        """Opens the associated file for reading"""
        return open(self.file)

    def parse(self, line):
        """Extracts ip range from given line"""
        raise NotImplementedError

    def read(self, callback):
        """Calls callback on each ip range in the file"""
        for start, end in self.readranges():
            try:
                callback(IP.parse(start), IP.parse(end))
            except BadIP as ex:
                log.error('Failed to parse IP: %s', ex)
        return self.file

    def is_ignored(self, line):
        """Ignore commented lines and blank lines"""
        line = line.strip()
        return line.startswith('#') or not line

    def is_valid(self):
        """Determines whether file is valid for this reader"""
        blocklist = self.open()
        valid = True
        for line in blocklist:
            line = decode_bytes(line)
            if not self.is_ignored(line):
                try:
                    (start, end) = self.parse(line)
                    if not re.match(r'^(\d{1,3}\.){4}$', start + '.') or not re.match(
                        r'^(\d{1,3}\.){4}$', end + '.'
                    ):
                        valid = False
                except Exception:
                    valid = False
                break
        blocklist.close()
        return valid

    @raises_errors_as(ReaderParseError)
    def readranges(self):
        """Yields each ip range from the file"""
        blocklist = self.open()
        for line in blocklist:
            line = decode_bytes(line)
            if not self.is_ignored(line):
                yield self.parse(line)
        blocklist.close()


class EmuleReader(BaseReader):
    """Blocklist reader for emule style blocklists"""

    def parse(self, line):
        return line.strip().split(' , ')[0].split(' - ')


class SafePeerReader(BaseReader):
    """Blocklist reader for SafePeer style blocklists"""

    def parse(self, line):
        return line.strip().split(':')[-1].split('-')


class PeerGuardianReader(SafePeerReader):
    """Blocklist reader for PeerGuardian style blocklists"""

    pass



#
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from gi.repository import Gtk

import deluge.component as component
from deluge.plugins.pluginbase import Gtk3PluginBase
from deluge.ui.client import client

from .common import get_resource

log = logging.getLogger(__name__)


class GtkUI(Gtk3PluginBase):
    def enable(self):
        self.builder = Gtk.Builder()
        self.builder.add_from_file(get_resource('config.ui'))

        component.get('Preferences').add_page(
            _('WebUi'), self.builder.get_object('prefs_box')
        )
        component.get('PluginManager').register_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').register_hook(
            'on_show_prefs', self.on_show_prefs
        )
        client.webui.get_config().addCallback(self.cb_get_config)
        client.webui.got_deluge_web().addCallback(self.cb_chk_deluge_web)

    def disable(self):
        component.get('Preferences').remove_page(_('WebUi'))
        component.get('PluginManager').deregister_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').deregister_hook(
            'on_show_prefs', self.on_show_prefs
        )

    def on_apply_prefs(self):
        if not self.have_web:
            return
        log.debug('applying prefs for WebUi')
        config = {
            'enabled': self.builder.get_object('enabled_checkbutton').get_active(),
            'ssl': self.builder.get_object('ssl_checkbutton').get_active(),
            'port': self.builder.get_object('port_spinbutton').get_value_as_int(),
        }
        client.webui.set_config(config)

    def on_show_prefs(self):
        client.webui.get_config().addCallback(self.cb_get_config)

    def cb_get_config(self, config):
        """Callback for on show_prefs."""
        self.builder.get_object('enabled_checkbutton').set_active(config['enabled'])
        self.builder.get_object('ssl_checkbutton').set_active(config['ssl'])
        self.builder.get_object('port_spinbutton').set_value(config['port'])

    def cb_chk_deluge_web(self, have_web):
        self.have_web = have_web
        if have_web:
            return
        self.builder.get_object('settings_vbox').set_sensitive(False)

        vbox = self.builder.get_object('prefs_box')

        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=0)
        icon = Gtk.Image.new_from_icon_name('dialog-error', Gtk.IconSize.BUTTON)
        icon.set_padding(5, 5)
        hbox.pack_start(icon, False, False, 0)

        label = Gtk.Label(
            _(
                'The Deluge web interface is not installed, '
                'please install the\ninterface and try again'
            )
        )
        label.set_alignment(0, 0.5)
        label.set_padding(5, 5)
        hbox.pack_start(label, False, False, 0)

        vbox.pack_start(hbox, False, False, 10)
        vbox.reorder_child(hbox, 0)
        vbox.show_all()



#
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from gi.repository.Gdk import ModifierType
from gi.repository.Gtk import SeparatorMenuItem, accel_groups_from_object
from gi.repository.Gtk.AccelFlags import VISIBLE

from deluge.configmanager import ConfigManager


def accel_swap(item, group, skey, smod, dkey, dmod):
    # Accel map hack broken, see ticket #3078
    # item.remove_accelerator(group, ord(skey), smod)
    item.add_accelerator('activate', group, ord(dkey), dmod, VISIBLE)


def accel_meta(item, group, key):
    accel_swap(item, group, key, ModifierType.CONTROL_MASK, key, ModifierType.META_MASK)


def menubar_osx(gtkui, osxapp):
    main_builder = gtkui.mainwindow.get_builder()
    menubar = main_builder.get_object('menubar')
    group = accel_groups_from_object(gtkui.mainwindow.window)[0]

    config = ConfigManager('gtk3ui.conf')

    # NOTE: accel maps doesn't work with glade file format
    # because of libglade not setting MenuItem accel groups
    # That's why we remove / set accelerators by hand... (dirty)
    # Clean solution: migrate glades files to gtkbuilder format
    file_menu = main_builder.get_object('menu_file').get_submenu()
    file_items = file_menu.get_children()
    accel_meta(file_items[0], group, 'o')
    accel_meta(file_items[1], group, 'n')
    quit_all_item = file_items[3]
    accel_swap(
        quit_all_item,
        group,
        'q',
        ModifierType.SHIFT_MASK | ModifierType.CONTROL_MASK,
        'q',
        ModifierType.SHIFT_MASK | ModifierType.META_MASK,
    )
    for item in range(2, len(file_items)):  # remove quits
        file_menu.remove(file_items[item])

    menu_widget = main_builder.get_object('menu_edit')
    edit_menu = menu_widget.get_submenu()
    edit_items = edit_menu.get_children()
    pref_item = edit_items[0]
    accel_swap(
        pref_item, group, 'p', ModifierType.CONTROL_MASK, ',', ModifierType.META_MASK
    )
    edit_menu.remove(pref_item)

    conn_item = edit_items[1]
    accel_meta(conn_item, group, 'm')
    edit_menu.remove(conn_item)

    menubar.remove(menu_widget)

    help_menu = main_builder.get_object('menu_help').get_submenu()
    help_items = help_menu.get_children()
    about_item = help_items[4]
    help_menu.remove(about_item)
    help_menu.remove(help_items[3])  # separator

    menubar.hide()
    osxapp.set_menu_bar(menubar)
    # populate app menu
    osxapp.insert_app_menu_item(about_item, 0)
    osxapp.insert_app_menu_item(SeparatorMenuItem(), 1)
    osxapp.insert_app_menu_item(pref_item, 2)
    if not config['standalone']:
        osxapp.insert_app_menu_item(conn_item, 3)
    if quit_all_item.get_visible():
        osxapp.insert_app_menu_item(SeparatorMenuItem(), 4)
        osxapp.insert_app_menu_item(quit_all_item, 5)



#
# Copyright (C) 2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
from datetime import datetime

import gi  # isort:skip (Required before Gtk import).

gi.require_version('Gtk', '3.0')

# isort:imports-thirdparty
from gi.repository import Gtk

# isort:imports-firstparty
import deluge.common
import deluge.component as component
from deluge.plugins.pluginbase import Gtk3PluginBase
from deluge.ui.client import client

# isort:imports-localfolder
from . import common

log = logging.getLogger(__name__)


class GtkUI(Gtk3PluginBase):
    def enable(self):
        log.debug('Blocklist GtkUI enable..')
        self.plugin = component.get('PluginManager')

        self.load_preferences_page()

        self.status_item = component.get('StatusBar').add_item(
            image=common.get_resource('blocklist16.png'),
            text='',
            callback=self._on_status_item_clicked,
            tooltip=_('Blocked IP Ranges /Whitelisted IP Ranges'),
        )

        # Register some hooks
        self.plugin.register_hook('on_apply_prefs', self._on_apply_prefs)
        self.plugin.register_hook('on_show_prefs', self._on_show_prefs)

    def disable(self):
        log.debug('Blocklist GtkUI disable..')

        # Remove the preferences page
        self.plugin.remove_preferences_page(_('Blocklist'))

        # Remove status item
        component.get('StatusBar').remove_item(self.status_item)
        del self.status_item

        # Deregister the hooks
        self.plugin.deregister_hook('on_apply_prefs', self._on_apply_prefs)
        self.plugin.deregister_hook('on_show_prefs', self._on_show_prefs)

        del self.glade

    def update(self):
        def _on_get_status(status):
            if status['state'] == 'Downloading':
                self.table_info.hide()
                self.builder.get_object('button_check_download').set_sensitive(False)
                self.builder.get_object('button_force_download').set_sensitive(False)
                self.builder.get_object('image_up_to_date').hide()

                self.status_item.set_text(
                    'Downloading %.2f%%' % (status['file_progress'] * 100)
                )
                self.progress_bar.set_text(
                    'Downloading %.2f%%' % (status['file_progress'] * 100)
                )
                self.progress_bar.set_fraction(status['file_progress'])
                self.progress_bar.show()

            elif status['state'] == 'Importing':
                self.table_info.hide()
                self.builder.get_object('button_check_download').set_sensitive(False)
                self.builder.get_object('button_force_download').set_sensitive(False)
                self.builder.get_object('image_up_to_date').hide()

                self.status_item.set_text('Importing ' + str(status['num_blocked']))
                self.progress_bar.set_text('Importing %s' % (status['num_blocked']))
                self.progress_bar.pulse()
                self.progress_bar.show()

            elif status['state'] == 'Idle':
                self.progress_bar.hide()
                self.builder.get_object('button_check_download').set_sensitive(True)
                self.builder.get_object('button_force_download').set_sensitive(True)
                if status['up_to_date']:
                    self.builder.get_object('image_up_to_date').show()
                else:
                    self.builder.get_object('image_up_to_date').hide()

                self.table_info.show()
                self.status_item.set_text('%(num_blocked)s/%(num_whited)s' % status)

                self.builder.get_object('label_filesize').set_text(
                    deluge.common.fsize(status['file_size'])
                )
                self.builder.get_object('label_modified').set_text(
                    datetime.fromtimestamp(status['file_date']).strftime('%c')
                )
                self.builder.get_object('label_type').set_text(status['file_type'])
                self.builder.get_object('label_url').set_text(status['file_url'])

        client.blocklist.get_status().addCallback(_on_get_status)

    def _on_show_prefs(self):
        def _on_get_config(config):
            log.trace('Loaded config: %s', config)
            self.builder.get_object('entry_url').set_text(config['url'])
            self.builder.get_object('spin_check_days').set_value(
                config['check_after_days']
            )
            self.builder.get_object('chk_import_on_start').set_active(
                config['load_on_start']
            )
            self.populate_whitelist(config['whitelisted'])

        client.blocklist.get_config().addCallback(_on_get_config)

    def _on_apply_prefs(self):
        config = {}
        config['url'] = self.builder.get_object('entry_url').get_text().strip()
        config['check_after_days'] = self.builder.get_object(
            'spin_check_days'
        ).get_value_as_int()
        config['load_on_start'] = self.builder.get_object(
            'chk_import_on_start'
        ).get_active()
        config['whitelisted'] = [
            ip[0] for ip in self.whitelist_model if ip[0] != 'IP HERE'
        ]
        client.blocklist.set_config(config)

    def _on_button_check_download_clicked(self, widget):
        self._on_apply_prefs()
        client.blocklist.check_import()

    def _on_button_force_download_clicked(self, widget):
        self._on_apply_prefs()
        client.blocklist.check_import(force=True)

    def _on_status_item_clicked(self, widget, event):
        component.get('Preferences').show(_('Blocklist'))

    def load_preferences_page(self):
        """Initializes the preferences page and adds it to the preferences dialog"""
        # Load the preferences page
        self.builder = Gtk.Builder()
        self.builder.add_from_file(common.get_resource('blocklist_pref.ui'))

        self.whitelist_frame = self.builder.get_object('whitelist_frame')
        self.progress_bar = self.builder.get_object('progressbar')
        self.table_info = self.builder.get_object('table_info')

        # Hide the progress bar initially
        self.progress_bar.hide()
        self.table_info.show()

        # Create the whitelisted model
        self.build_whitelist_model_treeview()

        self.builder.connect_signals(
            {
                'on_button_check_download_clicked': self._on_button_check_download_clicked,
                'on_button_force_download_clicked': self._on_button_force_download_clicked,
                'on_whitelist_add_clicked': (
                    self.on_add_button_clicked,
                    self.whitelist_treeview,
                ),
                'on_whitelist_remove_clicked': (
                    self.on_delete_button_clicked,
                    self.whitelist_treeview,
                ),
            }
        )

        # Set button icons
        self.builder.get_object('image_download').set_from_file(
            common.get_resource('blocklist_download24.png')
        )

        self.builder.get_object('image_import').set_from_file(
            common.get_resource('blocklist_import24.png')
        )

        # Update the preferences page with config values from the core
        self._on_show_prefs()

        # Add the page to the preferences dialog
        self.plugin.add_preferences_page(
            _('Blocklist'), self.builder.get_object('blocklist_prefs_box')
        )

    def build_whitelist_model_treeview(self):
        self.whitelist_treeview = self.builder.get_object('whitelist_treeview')
        treeview_selection = self.whitelist_treeview.get_selection()
        treeview_selection.connect(
            'changed', self.on_whitelist_treeview_selection_changed
        )
        self.whitelist_model = Gtk.ListStore(str, bool)
        renderer = Gtk.CellRendererText()
        renderer.connect('edited', self.on_cell_edited, self.whitelist_model)
        renderer.ip = 0

        column = Gtk.TreeViewColumn('IPs', renderer, text=0, editable=1)
        column.set_expand(True)
        self.whitelist_treeview.append_column(column)
        self.whitelist_treeview.set_model(self.whitelist_model)

    def on_cell_edited(self, cell, path_string, new_text, model):
        # iter = model.get_iter_from_string(path_string)
        # path = model.get_path(iter)[0]
        try:
            ip = common.IP.parse(new_text)
            model.set(model.get_iter_from_string(path_string), 0, ip.address)
        except common.BadIP as ex:
            model.remove(model.get_iter_from_string(path_string))
            from deluge.ui.gtkui import dialogs

            d = dialogs.ErrorDialog(_('Bad IP address'), ex.message)
            d.run()

    def on_whitelist_treeview_selection_changed(self, selection):
        model, selected_connection_iter = selection.get_selected()
        if selected_connection_iter:
            self.builder.get_object('whitelist_delete').set_property('sensitive', True)
        else:
            self.builder.get_object('whitelist_delete').set_property('sensitive', False)

    def on_add_button_clicked(self, widget, treeview):
        model = treeview.get_model()
        model.set(model.append(), 0, 'IP HERE', 1, True)

    def on_delete_button_clicked(self, widget, treeview):
        selection = treeview.get_selection()
        model, selected_iter = selection.get_selected()
        if selected_iter:
            # path = model.get_path(iter)[0]
            model.remove(selected_iter)

    def populate_whitelist(self, whitelist):
        self.whitelist_model.clear()
        for ip in whitelist:
            self.whitelist_model.set(self.whitelist_model.append(), 0, ip, 1, True)



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from os.path import sep as dirsep

import deluge.component as component
import deluge.ui.console.utils.colors as colors
from deluge.common import TORRENT_STATE, fsize, fspeed
from deluge.ui.client import client
from deluge.ui.common import FILE_PRIORITY
from deluge.ui.console.utils.format_utils import (
    f_progressbar,
    f_seedrank_dash,
    format_date_never,
    format_progress,
    format_time,
    ftotal_sized,
    pad_string,
    remove_formatting,
    shorten_hash,
    strwidth,
    trim_string,
)

from . import BaseCommand

STATUS_KEYS = [
    'state',
    'download_location',
    'tracker_host',
    'tracker_status',
    'next_announce',
    'name',
    'total_size',
    'progress',
    'num_seeds',
    'total_seeds',
    'num_peers',
    'total_peers',
    'eta',
    'download_payload_rate',
    'upload_payload_rate',
    'ratio',
    'distributed_copies',
    'num_pieces',
    'piece_length',
    'total_done',
    'files',
    'file_priorities',
    'file_progress',
    'peers',
    'is_seed',
    'is_finished',
    'active_time',
    'seeding_time',
    'time_since_transfer',
    'last_seen_complete',
    'seed_rank',
    'all_time_download',
    'total_uploaded',
    'total_payload_download',
    'total_payload_upload',
    'time_added',
    'label',
]

# Add filter specific state to torrent states
STATES = ['Active'] + TORRENT_STATE


class Command(BaseCommand):
    """Show information about the torrents"""

    sort_help = 'sort items.  Possible keys: ' + ', '.join(STATUS_KEYS)

    epilog = """
  You can give the first few characters of a torrent-id to identify the torrent.

  Tab Completion in interactive mode (info *pattern*<tab>):\n
      | First press of <tab> will output up to 15 matches;
      | hitting <tab> a second time, will print 15 more matches;
      | and a third press will print all remaining matches.
      | (To modify behaviour of third <tab>, set `third_tab_lists_all` to False)
"""

    def add_arguments(self, parser):
        parser.add_argument(
            '-v',
            '--verbose',
            action='store_true',
            default=False,
            dest='verbose',
            help=_('Show more information per torrent.'),
        )
        parser.add_argument(
            '-d',
            '--detailed',
            action='store_true',
            default=False,
            dest='detailed',
            help=_('Show more detailed information including files and peers.'),
        )
        parser.add_argument(
            '-s',
            '--state',
            action='store',
            dest='state',
            help=_('Show torrents with state STATE: %s.' % (', '.join(STATES))),
        )
        parser.add_argument(
            '--sort',
            action='store',
            type=str,
            default='',
            dest='sort',
            help=self.sort_help,
        )
        parser.add_argument(
            '--sort-reverse',
            action='store',
            type=str,
            default='',
            dest='sort_rev',
            help=_('Same as --sort but items are in reverse order.'),
        )
        parser.add_argument(
            'torrent_ids',
            metavar='<torrent-id>',
            nargs='*',
            help=_('One or more torrent ids. If none is given, list all'),
        )

    def add_subparser(self, subparsers):
        parser = subparsers.add_parser(
            self.name,
            prog=self.name,
            help=self.__doc__,
            description=self.__doc__,
            epilog=self.epilog,
        )
        self.add_arguments(parser)

    def handle(self, options):
        self.console = component.get('ConsoleUI')
        # Compile a list of torrent_ids to request the status of
        torrent_ids = []

        if options.torrent_ids:
            for t_id in options.torrent_ids:
                torrent_ids.extend(self.console.match_torrent(t_id))
        else:
            torrent_ids.extend(self.console.match_torrent(''))

        def on_torrents_status(status):
            # Print out the information for each torrent
            sort_key = options.sort
            sort_reverse = False
            if not sort_key:
                sort_key = options.sort_rev
                sort_reverse = True
            if not sort_key:
                sort_key = 'name'
                sort_reverse = False
            if sort_key not in STATUS_KEYS:
                self.console.write('')
                self.console.write(
                    '{!error!}Unknown sort key: ' + sort_key + ', will sort on name'
                )
                sort_key = 'name'
                sort_reverse = False
            for key, value in sorted(
                status.items(),
                key=lambda x: x[1].get(sort_key),
                reverse=sort_reverse,
            ):
                self.show_info(key, status[key], options.verbose, options.detailed)

        def on_torrents_status_fail(reason):
            self.console.write('{!error!}Error getting torrent info: %s' % reason)

        status_dict = {'id': torrent_ids}

        if options.state:
            options.state = options.state.capitalize()
            if options.state in STATES:
                status_dict.state = options.state
            else:
                self.console.write('Invalid state: %s' % options.state)
                self.console.write('Possible values are: %s.' % (', '.join(STATES)))
                return

        d = client.core.get_torrents_status(status_dict, STATUS_KEYS)
        d.addCallback(on_torrents_status)
        d.addErrback(on_torrents_status_fail)
        return d

    def show_file_info(self, torrent_id, status):
        spaces_per_level = 2

        if hasattr(self.console, 'screen'):
            cols = self.console.screen.cols
        else:
            cols = 80

        prevpath = []
        for index, torrent_file in enumerate(status['files']):
            filename = torrent_file['path'].split(dirsep)[-1]
            filepath = torrent_file['path'].split(dirsep)[:-1]

            for depth, subdir in enumerate(filepath):
                indent = ' ' * depth * spaces_per_level
                if depth >= len(prevpath):
                    self.console.write(f'{indent}{{!cyan!}}{subdir}')
                elif subdir != prevpath[depth]:
                    self.console.write(f'{indent}{{!cyan!}}{subdir}')

            depth = len(filepath)

            indent = ' ' * depth * spaces_per_level

            col_filename = indent + filename
            col_size = ' ({!cyan!}%s{!input!})' % fsize(torrent_file['size'])
            col_progress = ' {!input!}%.2f%%' % (status['file_progress'][index] * 100)

            col_priority = ' {!info!}Priority: '

            file_priority = FILE_PRIORITY[status['file_priorities'][index]]

            if status['file_progress'][index] != 1.0:
                if file_priority == 'Skip':
                    col_priority += '{!error!}'
                else:
                    col_priority += '{!success!}'
            else:
                col_priority += '{!input!}'
            col_priority += file_priority

            def tlen(string):
                return strwidth(remove_formatting(string))

            col_all_info = col_size + col_progress + col_priority
            # Check how much space we've got left after writing all the info
            space_left = cols - tlen(col_all_info)
            # And how much we will potentially have with the longest possible column
            maxlen_space_left = cols - tlen(' (1000.0 MiB) 100.00% Priority: Normal')
            if maxlen_space_left > tlen(col_filename) + 1:
                # If there is enough space, pad it all nicely
                col_all_info = ''
                col_all_info += ' ('
                spaces_to_add = tlen(' (1000.0 MiB)') - tlen(col_size)
                col_all_info += ' ' * spaces_to_add
                col_all_info += col_size[2:]
                spaces_to_add = tlen(' 100.00%') - tlen(col_progress)
                col_all_info += ' ' * spaces_to_add
                col_all_info += col_progress
                spaces_to_add = tlen(' Priority: Normal') - tlen(col_priority)
                col_all_info += col_priority
                col_all_info += ' ' * spaces_to_add
                # And remember to put it to the left!
                col_filename = pad_string(
                    col_filename, maxlen_space_left - 2, side='right'
                )
            elif space_left > tlen(col_filename) + 1:
                # If there is enough space, put the info to the right
                col_filename = pad_string(col_filename, space_left - 2, side='right')
            else:
                # And if there is not, shorten the name
                col_filename = trim_string(col_filename, space_left, True)
            self.console.write(col_filename + col_all_info)

            prevpath = filepath

    def show_peer_info(self, torrent_id, status):
        if len(status['peers']) == 0:
            self.console.write('    None')
        else:
            s = ''
            for peer in status['peers']:
                if peer['seed']:
                    s += '%sSeed\t{!input!}' % colors.state_color['Seeding']
                else:
                    s += '%sPeer\t{!input!}' % colors.state_color['Downloading']

                s += peer['country'] + '\t'

                if peer['ip'].count(':') == 1:
                    # IPv4
                    s += peer['ip']
                else:
                    # IPv6
                    s += '[{}]:{}'.format(
                        ':'.join(peer['ip'].split(':')[:-1]),
                        peer['ip'].split(':')[-1],
                    )

                c = peer['client']
                s += '\t' + c

                if len(c) < 16:
                    s += '\t\t'
                else:
                    s += '\t'
                s += '{}{}\t{}{}'.format(
                    colors.state_color['Seeding'],
                    fspeed(peer['up_speed']),
                    colors.state_color['Downloading'],
                    fspeed(peer['down_speed']),
                )
                s += '\n'

            self.console.write(s[:-1])

    def show_info(self, torrent_id, status, verbose=False, detailed=False):
        """
        Writes out the torrents information to the screen.

        Format depends on switches given.
        """
        self.console.set_batch_write(True)

        if hasattr(self.console, 'screen'):
            cols = self.console.screen.cols
        else:
            cols = 80

        sep = ' '

        if verbose or detailed:
            self.console.write('{!info!}Name: {!input!}%s' % (status['name']))
            self.console.write('{!info!}ID: {!input!}%s' % (torrent_id))
            s = '{{!info!}}State: {}{}'.format(
                colors.state_color[status['state']],
                status['state'],
            )
            # Only show speed if active
            if status['state'] in ('Seeding', 'Downloading'):
                if status['state'] != 'Seeding':
                    s += sep
                    s += '{!info!}Down Speed: {!input!}%s' % fspeed(
                        status['download_payload_rate'], shortform=True
                    )
                s += sep
                s += '{!info!}Up Speed: {!input!}%s' % fspeed(
                    status['upload_payload_rate'], shortform=True
                )
            self.console.write(s)

            if status['state'] in ('Seeding', 'Downloading', 'Queued'):
                s = '{{!info!}}Seeds: {{!input!}}{} ({})'.format(
                    status['num_seeds'],
                    status['total_seeds'],
                )
                s += sep
                s += '{{!info!}}Peers: {{!input!}}{} ({})'.format(
                    status['num_peers'],
                    status['total_peers'],
                )
                s += sep
                s += (
                    '{!info!}Availability: {!input!}%.2f' % status['distributed_copies']
                )
                s += sep
                s += '{!info!}Seed Rank: {!input!}%s' % f_seedrank_dash(
                    status['seed_rank'], status['seeding_time']
                )
                self.console.write(s)

            total_done = fsize(status['total_done'], shortform=True)
            total_size = fsize(status['total_size'], shortform=True)
            if total_done == total_size:
                s = '{!info!}Size: {!input!}%s' % (total_size)
            else:
                s = f'{{!info!}}Size: {{!input!}}{total_done}/{total_size}'
            s += sep
            s += '{!info!}Downloaded: {!input!}%s' % fsize(
                status['all_time_download'], shortform=True
            )
            s += sep
            s += '{!info!}Uploaded: {!input!}%s' % fsize(
                status['total_uploaded'], shortform=True
            )
            s += sep
            s += '{!info!}Share Ratio: {!input!}%.2f' % status['ratio']
            self.console.write(s)

            s = '{!info!}ETA: {!input!}%s' % format_time(status['eta'])
            s += sep
            s += '{!info!}Seeding: {!input!}%s' % format_time(status['seeding_time'])
            s += sep
            s += '{!info!}Active: {!input!}%s' % format_time(status['active_time'])
            self.console.write(s)

            s = '{!info!}Last Transfer: {!input!}%s' % format_time(
                status['time_since_transfer']
            )
            s += sep
            s += '{!info!}Complete Seen: {!input!}%s' % format_date_never(
                status['last_seen_complete']
            )
            self.console.write(s)

            s = '{!info!}Tracker: {!input!}%s' % status['tracker_host']
            self.console.write(s)

            self.console.write(
                '{!info!}Tracker status: {!input!}%s' % status['tracker_status']
            )

            if not status['is_finished']:
                pbar = f_progressbar(
                    status['progress'], cols - (13 + len('%.2f%%' % status['progress']))
                )
                s = '{{!info!}}Progress: {{!input!}}{:.2f}% {}'.format(
                    status['progress'], pbar
                )
                self.console.write(s)

            s = '{!info!}Download Folder: {!input!}%s' % status['download_location']
            self.console.write(s)

            if 'label' in status:
                s = '{!info!}Label: {!input!}%s' % status['label']
                self.console.write(s)

            if detailed:
                self.console.write('\n{!info!}Files in torrent')
                self.show_file_info(torrent_id, status)
                self.console.write('{!info!}Connected peers')
                self.show_peer_info(torrent_id, status)
        else:
            up_color = colors.state_color['Seeding']
            down_color = colors.state_color['Downloading']

            s = '{}{}'.format(
                colors.state_color[status['state']],
                '[' + status['state'][0] + ']',
            )

            s += ' {!info!}' + format_progress(status['progress']).rjust(6, ' ')
            s += ' {!input!}%s' % (status['name'])

            # Shorten the ID if it's necessary. Pretty hacky
            # XXX: should make a nice function for it that can partition and shorten stuff
            space_left = cols - strwidth('[S] 99.99% ' + status['name'])

            if self.console.interactive and space_left >= len(sep + torrent_id):
                # Not enough line space so shorten the hash (for interactive mode).
                torrent_id = shorten_hash(torrent_id, space_left)
            s += sep
            s += '{!cyan!}%s' % torrent_id
            self.console.write(s)

            dl_info = '{!info!}DL: {!input!}'
            dl_info += '%s' % ftotal_sized(
                status['all_time_download'], status['total_payload_download']
            )

            if status['download_payload_rate'] > 0:
                dl_info += ' @ {}{}'.format(
                    down_color,
                    fspeed(status['download_payload_rate'], shortform=True),
                )

            ul_info = ' {!info!}UL: {!input!}'
            ul_info += '%s' % ftotal_sized(
                status['total_uploaded'], status['total_payload_upload']
            )
            if status['upload_payload_rate'] > 0:
                ul_info += ' @ {}{}'.format(
                    up_color,
                    fspeed(status['upload_payload_rate'], shortform=True),
                )

            eta = ' {!info!}ETA: {!magenta!}%s' % format_time(status['eta'])

            self.console.write('    ' + dl_info + ul_info + eta + '\n')

        self.console.set_batch_write(False)

    def complete(self, line):
        # We use the ConsoleUI torrent tab complete method
        return component.get('ConsoleUI').tab_complete_torrent(line)



#!/usr/bin/env python
#
# Copyright (C) 2013 Bro <bro.development@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
import warnings

from gi.repository import Gdk, GObject, Gtk
from gi.repository.GObject import SignalFlags

from deluge.common import resource_filename
from deluge.path_chooser_common import get_completion_paths

# Filter the pygobject signal warning:
#     g_value_get_int: assertion 'G_VALUE_HOLDS_INT (value)' failed.
# See: https://gitlab.gnome.org/GNOME/pygobject/issues/12
warnings.filterwarnings('ignore', '.*g_value_get_int.*G_VALUE_HOLDS_INT.*', Warning)


def is_ascii_value(keyval, ascii_key):
    try:
        # Set show/hide hidden files
        if chr(keyval) == ascii_key:
            return True
    except ValueError:
        # Not in ascii range
        pass
    return False


def key_is_up(keyval):
    return keyval == Gdk.KEY_Up or keyval == Gdk.KEY_KP_Up


def key_is_down(keyval):
    return keyval == Gdk.KEY_Down or keyval == Gdk.KEY_KP_Down


def key_is_up_or_down(keyval):
    return key_is_up(keyval) or key_is_down(keyval)


def key_is_pgup_or_pgdown(keyval):
    return keyval == Gdk.KEY_Page_Down or keyval == Gdk.KEY_Page_Up


def key_is_enter(keyval):
    return keyval == Gdk.KEY_Return or keyval == Gdk.KEY_KP_Enter


def path_without_trailing_path_sep(path):
    while path.endswith('/') or path.endswith('\\'):
        if path == '/':
            return path
        path = path[0:-1]
    return path


class ValueList:

    paths_without_trailing_path_sep = False

    def get_values_count(self):
        return len(self.tree_store)

    def get_values(self):
        """
        Returns the values in the list.
        """
        values = []
        for row in self.tree_store:
            values.append(row[0])
        return values

    def add_values(
        self, paths, append=True, scroll_to_row=False, clear=False, emit_signal=False
    ):
        """
        Add paths to the liststore

        :param paths: the paths to add
        :type  paths: list
        :param append: if the values should be appended or inserted
        :type  append: boolean
        :param scroll_to_row: if the treeview should scroll to the new row
        :type  scroll_to_row: boolean

        """
        if clear:
            self.tree_store.clear()

        for path in paths:
            if self.paths_without_trailing_path_sep:
                path = path_without_trailing_path_sep(path)
            if append:
                tree_iter = self.tree_store.append([path])
            else:
                tree_iter = self.tree_store.insert(0, [path])

            if scroll_to_row:
                self.treeview.grab_focus()
                tree_path = self.tree_store.get_path(tree_iter)
                # Scroll to path
                self.handle_list_scroll(path=tree_path)

        if emit_signal:
            self.emit('list-value-added', paths)
            self.emit('list-values-changed', self.get_values())

    def set_values(self, paths, scroll_to_row=False, preserve_selection=True):
        """
        Add paths to the liststore

        :param paths: the paths to add
        :type  paths: list
        :param scroll_to_row: if the treeview should scroll to the new row
        :type  scroll_to_row: boolean

        """
        if not (isinstance(paths, list) or isinstance(paths, tuple)):
            return
        sel = None
        if preserve_selection:
            sel = self.get_selection_path()
        self.add_values(paths, scroll_to_row=scroll_to_row, clear=True)
        if sel:
            self.treeview.get_selection().select_path(sel)

    def get_selection_path(self):
        """Returns the (first) selected path from a treeview"""
        tree_selection = self.treeview.get_selection()
        model, tree_paths = tree_selection.get_selected_rows()
        if len(tree_paths) > 0:
            return tree_paths[0]
        return None

    def get_selected_value(self):
        path = self.get_selection_path()
        if path:
            return self.tree_store[path][0]
        return None

    def remove_selected_path(self):
        path = self.get_selection_path()
        if path:
            path_value = self.tree_store[path][0]
            del self.tree_store[path]
            index = path[0]
            # The last row was deleted
            if index == len(self.tree_store):
                index -= 1
            if index >= 0:
                path = (index,)
            self.treeview.set_cursor(path)
            self.set_path_selected(path)
            self.emit('list-value-removed', path_value)
            self.emit('list-values-changed', self.get_values())

    def set_selected_value(self, value, select_first=False):
        """
        Select the row of the list with value

        :param value: the value to be selected
        :type  value: str
        :param select_first: if the first item should be selected if the value if not found.
        :type  select_first: boolean

        """
        for i, row in enumerate(self.tree_store):
            if row[0] == value:
                self.treeview.set_cursor(i)
                return
        # The value was not found
        if select_first:
            self.treeview.set_cursor((0,))
        else:
            self.treeview.get_selection().unselect_all()

    def set_path_selected(self, path):
        self.treeview.get_selection().select_path(path)

    def on_value_list_treeview_key_press_event(self, widget, event):
        """
        Mimics Combobox behavior

        Escape or Alt+Up: Close
        Enter or Return : Select
        """
        keyval = event.keyval
        state = event.get_state() & Gtk.accelerator_get_default_mod_mask()
        alt_up = (state == Gdk.ModifierType.MOD1_MASK) and key_is_up(keyval)
        if keyval == Gdk.KEY_Escape or alt_up:
            self.popdown()
            return True
        # Set entry value to the selected row
        elif key_is_enter(keyval):
            path = self.get_selection_path()
            if path:
                self.set_entry_value(path, popdown=True)
            return True
        return False

    def on_treeview_mouse_button_press_event(self, treeview, event, double_click=False):
        """
        When left clicking twice, the row value is set for the text entry
        and the popup is closed.

        """
        # This is left click
        if event.button != 3:
            # Double clicked a row, set this as the entry value
            # and close the popup
            if (double_click and event.type == Gdk.EventType._2BUTTON_PRESS) or (
                not double_click and event.type == Gdk.EventType.BUTTON_PRESS
            ):
                path = self.get_selection_path()
                if path:
                    self.set_entry_value(path, popdown=True)
                    return True
        return False

    def handle_list_scroll(
        self, _next=None, path=None, set_entry=False, swap=False, scroll_window=False
    ):
        """
        Handles changes to the row selection.

        :param _next: the direction to change selection. True means down and False means up.
            None means no change.
        :type  _next: boolean/None
        :param path: the current path. If None, the currently selected path is used.
        :type  path: tuple
        :param set_entry: if the new value should be set in the text entry.
        :type  set_entry: boolean
        :param swap: if the old and new value should be swapped
        :type  swap: boolean

        """
        if scroll_window:
            adjustment = self.completion_scrolled_window.get_vadjustment()

            visible_rows_height = self.get_values_count()
            if visible_rows_height > self.max_visible_rows:
                visible_rows_height = self.max_visible_rows

            visible_rows_height *= self.row_height
            value = adjustment.get_value()

            # Max adjustment value
            max_value = adjustment.get_upper() - visible_rows_height
            # Set adjustment increment to 3 times the row height
            adjustment.set_step_increment(self.row_height * 3)

            if _next:
                # If number of values is less than max rows, no scroll
                if self.get_values_count() < self.max_visible_rows:
                    return
                value += adjustment.get_step_increment()
                if value > max_value:
                    value = max_value
            else:
                value -= adjustment.get_step_increment()
                if value < 0:
                    value = 0
            adjustment.set_value(value)
            return

        if path is None:
            path = self.get_selection_path()
            if not path:
                # These options require a selected path
                if set_entry or swap:
                    return
                # This is a regular scroll, not setting value in entry or swapping rows,
                # so we find a path value anyways
                path = (0,)
                cursor = self.treeview.get_cursor()
                if cursor is not None and cursor[0] is not None:
                    path = cursor[0]
                else:
                    # Since cursor is none, we won't advance the index
                    _next = None

        # If _next is None, we won't change the selection
        if _next is not None:
            # We move the selection either one up or down.
            # If we reach end of list, we wrap
            index = path[0] if path else 0
            index = index + 1 if _next else index - 1
            if index >= len(self.tree_store):
                index = 0
            elif index < 0:
                index = len(self.tree_store) - 1

            # We have the index for the new path
            new_path = index
            if swap:
                p1 = self.tree_store[path][0]
                p2 = self.tree_store[new_path][0]
                self.tree_store.swap(
                    self.tree_store.get_iter(path), self.tree_store.get_iter(new_path)
                )
                self.emit('list-values-reordered', [p1, p2])
                self.emit('list-values-changed', self.get_values())
            path = new_path

        self.treeview.set_cursor(path)
        self.treeview.get_selection().select_path(path)
        if set_entry:
            self.set_entry_value(path)


class StoredValuesList(ValueList):
    def __init__(self):
        self.tree_store = self.builder.get_object('stored_values_tree_store')
        self.tree_column = self.builder.get_object('stored_values_treeview_column')
        self.rendererText = self.builder.get_object('stored_values_cellrenderertext')
        self.paths_without_trailing_path_sep = False

        # Add signal handlers
        self.signal_handlers[
            'on_stored_values_treeview_mouse_button_press_event'
        ] = self.on_treeview_mouse_button_press_event

        self.signal_handlers[
            'on_stored_values_treeview_key_press_event'
        ] = self.on_stored_values_treeview_key_press_event
        self.signal_handlers[
            'on_stored_values_treeview_key_release_event'
        ] = self.on_stored_values_treeview_key_release_event

        self.signal_handlers[
            'on_cellrenderertext_edited'
        ] = self.on_cellrenderertext_edited

    def on_cellrenderertext_edited(self, cellrenderertext, path, new_text):
        """
        Callback on the 'edited' signal.

        Sets the new text in the path and disables editing on the renderer.
        """
        new_text = path_without_trailing_path_sep(new_text)
        self.tree_store[path][0] = new_text
        self.rendererText.set_property('editable', False)

    def on_edit_path(self, path, column):
        """
        Starts editing on the provided path

        :param path: the paths to edit
        :type  path: tuple
        :param column: the column to edit
        :type  column: Gtk.TreeViewColumn

        """
        self.rendererText.set_property('editable', True)
        self.treeview.grab_focus()
        self.treeview.set_cursor(path, column=column, start_editing=True)

    def on_treeview_mouse_button_press_event(self, treeview, event):
        """
        Shows popup on selected row when right clicking
        When left clicking twice, the row value is set for the text entry
        and the popup is closed.

        """
        # This is left click
        if event.button != 3:
            super().on_treeview_mouse_button_press_event(
                treeview, event, double_click=True
            )
            return False

        # This is right click, create popup menu for this row
        x = int(event.x)
        y = int(event.y)
        time = event.time
        pthinfo = treeview.get_path_at_pos(x, y)
        if pthinfo is not None:
            path, col, cellx, celly = pthinfo
            treeview.grab_focus()
            treeview.set_cursor(path, col, 0)

            self.path_list_popup = Gtk.Menu()
            menuitem_edit = Gtk.MenuItem.new_with_label(_('Edit path'))
            self.path_list_popup.append(menuitem_edit)
            menuitem_remove = Gtk.MenuItem.new_with_label(_('Remove path'))
            self.path_list_popup.append(menuitem_remove)

            def on_edit_clicked(widget, path):
                self.on_edit_path(path, self.tree_column)

            def on_remove_clicked(widget, path):
                self.remove_selected_path()

            menuitem_edit.connect('activate', on_edit_clicked, path)
            menuitem_remove.connect('activate', on_remove_clicked, path)
            self.path_list_popup.popup(None, None, None, path, event.button, time)
            self.path_list_popup.show_all()

    def remove_selected_path(self):
        ValueList.remove_selected_path(self)
        # Resize popup
        PathChooserPopup.popup(self)

    def on_stored_values_treeview_key_press_event(self, widget, event):
        super().on_value_list_treeview_key_press_event(widget, event)
        # Prevent the default event handler to move the cursor in the list
        if key_is_up_or_down(event.keyval):
            return True

    def on_stored_values_treeview_key_release_event(self, widget, event):
        """
        Mimics Combobox behavior

        Escape or Alt+Up: Close
        Enter or Return : Select

        """
        keyval = event.keyval
        ctrl = event.get_state() & Gdk.ModifierType.CONTROL_MASK

        # Edit selected row
        if keyval in [Gdk.KEY_Left, Gdk.KEY_Right, Gdk.KEY_space]:
            path = self.get_selection_path()
            if path:
                self.on_edit_path(path, self.tree_column)
        elif key_is_up_or_down(keyval):
            # Swap the row value
            if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
                self.handle_list_scroll(_next=key_is_down(keyval), swap=True)
            else:
                self.handle_list_scroll(_next=key_is_down(keyval))
        elif key_is_pgup_or_pgdown(event.keyval):
            # The cursor has been changed by the default key-press-event handler
            # so set the path of the cursor selected
            self.set_path_selected(self.treeview.get_cursor()[0])
        elif ctrl:
            # Handle key bindings for manipulating the list
            # Remove the selected entry
            if is_ascii_value(keyval, 'r'):
                self.remove_selected_path()
                return True
            # Add current value to saved list
            elif is_ascii_value(keyval, 's'):
                super(
                    PathChooserComboBox, self
                ).add_current_value_to_saved_list()  # pylint: disable=bad-super-call
                return True
            # Edit selected value
            elif is_ascii_value(keyval, 'e'):
                self.edit_selected_path()
                return True


class CompletionList(ValueList):
    def __init__(self):
        self.tree_store = self.builder.get_object('completion_tree_store')
        self.tree_column = self.builder.get_object('completion_treeview_column')
        self.rendererText = self.builder.get_object('completion_cellrenderertext')
        self.completion_scrolled_window = self.builder.get_object(
            'completion_scrolled_window'
        )
        self.signal_handlers[
            'on_completion_treeview_key_press_event'
        ] = self.on_completion_treeview_key_press_event
        self.signal_handlers[
            'on_completion_treeview_motion_notify_event'
        ] = self.on_completion_treeview_motion_notify_event

        # Add super class signal handler
        self.signal_handlers[
            'on_completion_treeview_mouse_button_press_event'
        ] = super().on_treeview_mouse_button_press_event

    def reduce_values(self, prefix):
        """
        Reduce the values in the liststore to those starting with the prefix.

        :param prefix: the prefix to be matched
        :type  paths: string

        """
        values = self.get_values()
        matching_values = []
        for v in values:
            if v.startswith(prefix):
                matching_values.append(v)
        self.add_values(matching_values, clear=True)

    def on_completion_treeview_key_press_event(self, widget, event):
        ret = super().on_value_list_treeview_key_press_event(widget, event)
        if ret:
            return ret
        keyval = event.keyval
        ctrl = event.get_state() & Gdk.ModifierType.CONTROL_MASK
        if key_is_up_or_down(keyval):
            self.handle_list_scroll(_next=key_is_down(keyval))
            return True
        elif ctrl:
            # Set show/hide hidden files
            if is_ascii_value(keyval, 'h'):
                self.path_entry.set_show_hidden_files(
                    not self.path_entry.get_show_hidden_files(), do_completion=True
                )
                return True

    def on_completion_treeview_motion_notify_event(self, widget, event):
        if event.is_hint:
            x, y, state = event.window.get_pointer()
        else:
            x = event.x
            y = event.y

        path = self.treeview.get_path_at_pos(int(x), int(y))
        if path:
            self.handle_list_scroll(path=path[0], _next=None)


class PathChooserPopup:
    """This creates the popop window for the ComboEntry."""

    def __init__(self, min_visible_rows, max_visible_rows, popup_alignment_widget):
        self.min_visible_rows = min_visible_rows
        # Maximum number of rows to display without scrolling
        self.set_max_popup_rows(max_visible_rows)
        self.popup_window.realize()
        self.alignment_widget = popup_alignment_widget
        # If set, the height of this widget is the minimum height
        self.popup_buttonbox = None

    def popup(self):
        """Make the popup visible."""
        # Entry is not yet visible
        if not self.path_entry.get_realized():
            return
        self.set_window_position_and_size()

    def popdown(self):
        if not self.is_popped_up():
            return
        if not self.path_entry.get_realized():
            return
        self.popup_window.grab_remove()
        self.popup_window.hide()

    def is_popped_up(self):
        """Check if window is popped up.

        Returns:
            bool: True if popped up, False otherwise.

        """
        return self.popup_window.get_mapped()

    def set_window_position_and_size(self):
        if len(self.tree_store) < self.min_visible_rows:
            return False
        x, y, width, height = self.get_position()
        self.popup_window.set_size_request(width, height)
        self.popup_window.resize(width, height)
        self.popup_window.move(x, y)
        self.popup_window.show_all()

    def get_position(self):
        """
        Returns the size of the popup window and the coordinates on the screen.

        """

        # Necessary for the first call, to make treeview.size_request give sensible values
        # self.popup_window.realize()
        self.treeview.realize()

        # We start with the coordinates of the parent window
        z, x, y = self.path_entry.get_window().get_origin()

        # Add the position of the alignment_widget relative to the parent window.
        x += self.alignment_widget.get_allocation().x
        y += self.alignment_widget.get_allocation().y

        height_extra = 8
        buttonbox_width = 0
        height = self.popup_window.get_preferred_height()[1]
        width = self.popup_window.get_preferred_width()[1]

        if self.popup_buttonbox:
            buttonbox_height = max(
                self.popup_buttonbox.get_preferred_height()[1],
                self.popup_buttonbox.get_allocation().height,
            )
            buttonbox_width = max(
                self.popup_buttonbox.get_preferred_width()[1],
                self.popup_buttonbox.get_allocation().width,
            )
            treeview_width = self.treeview.get_preferred_width()[1]
            # After removing an element from the tree store, self.treeview.get_preferred_width()[0]
            # returns -1 for some reason, so the requested width cannot be used until the treeview
            # has been displayed once.
            if treeview_width != -1:
                width = treeview_width + buttonbox_width
            # The list is empty, so ignore initial popup width request
            # Will be set to the minimum width next
            elif len(self.tree_store) == 0:
                width = 0

        if width < self.alignment_widget.get_allocation().width:
            width = self.alignment_widget.get_allocation().width

        # 10 is extra spacing
        content_width = self.treeview.get_preferred_width()[1] + buttonbox_width + 10

        # Adjust height according to number of list items
        if len(self.tree_store) > 0 and self.max_visible_rows > 0:
            # The height for one row in the list
            self.row_height = self.treeview.get_preferred_height()[1] / len(
                self.tree_store
            )
            # Set height to number of rows
            height = len(self.tree_store) * self.row_height + height_extra
            # Adjust the height according to the max number of rows
            max_height = self.row_height * self.max_visible_rows
            # Restrict height to max_visible_rows
            if max_height + height_extra < height:
                height = max_height
                height += height_extra
                # Increase width because of vertical scrollbar
                content_width += 15

        if self.popup_buttonbox:
            # Minimum height is the height of the button box
            if height < buttonbox_height + height_extra:
                height = buttonbox_height + height_extra

        if content_width > width:
            width = content_width

        screen = self.path_entry.get_screen()
        monitor_num = screen.get_monitor_at_window(self.path_entry.get_window())
        monitor = screen.get_monitor_geometry(monitor_num)

        if x < monitor.x:
            x = monitor.x
        elif x + width > monitor.x + monitor.width:
            x = monitor.x + monitor.width - width

        # Set the position
        if (
            y + self.path_entry.get_allocation().height + height
            <= monitor.y + monitor.height
        ):
            y += self.path_entry.get_allocation().height
        # Not enough space downwards on the screen
        elif y - height >= monitor.y:
            y -= height
        elif (
            monitor.y + monitor.height - (y + self.path_entry.get_allocation().height)
            > y - monitor.y
        ):
            y += self.path_entry.get_allocation().height
            height = monitor.y + monitor.height - y
        else:
            height = y - monitor.y
            y = monitor.y

        return x, y, width, height

    def popup_grab_window(self):
        activate_time = 0
        if (
            Gdk.pointer_grab(
                self.popup_window.get_window(),
                True,
                (
                    Gdk.EventMask.BUTTON_PRESS_MASK
                    | Gdk.EventMask.BUTTON_RELEASE_MASK
                    | Gdk.EventMask.POINTER_MOTION_MASK
                ),
                None,
                None,
                activate_time,
            )
            == 0
        ):
            if (
                Gdk.keyboard_grab(self.popup_window.get_window(), True, activate_time)
                == 0
            ):
                return True
            else:
                self.popup_window.get_window().get_display().pointer_ungrab(
                    activate_time
                )
                return False
        return False

    def set_entry_value(self, path, popdown=False):
        """

        Sets the text of the entry to the value in path
        """
        self.path_entry.set_text(
            self.tree_store[path][0], set_file_chooser_folder=True, trigger_event=True
        )
        if popdown:
            self.popdown()

    def set_max_popup_rows(self, rows):
        try:
            int(rows)
        except Exception:
            self.max_visible_rows = 20
            return
        self.max_visible_rows = rows

    def get_max_popup_rows(self):
        return self.max_visible_rows

    #################
    # Callbacks
    #################

    def on_popup_window_button_press_event(self, window, event):
        # If we're clicking outside of the window close the popup
        allocation = self.popup_window.get_allocation()

        if (event.x < allocation.x or event.x > allocation.width) or (
            event.y < allocation.y or event.y > allocation.height
        ):
            self.popdown()


class StoredValuesPopup(StoredValuesList, PathChooserPopup):
    """

    The stored values popup

    """

    def __init__(self, builder, path_entry, max_visible_rows, popup_alignment_widget):
        self.builder = builder
        self.treeview = self.builder.get_object('stored_values_treeview')
        self.popup_window = self.builder.get_object('stored_values_popup_window')
        self.button_default = self.builder.get_object('button_default')
        self.path_entry = path_entry
        self.text_entry = path_entry.text_entry

        self.signal_handlers = {}
        PathChooserPopup.__init__(self, 0, max_visible_rows, popup_alignment_widget)
        StoredValuesList.__init__(self)

        self.popup_buttonbox = self.builder.get_object('buttonbox')

        # Add signal handlers
        self.signal_handlers[
            'on_buttonbox_key_press_event'
        ] = self.on_buttonbox_key_press_event
        self.signal_handlers[
            'on_stored_values_treeview_scroll_event'
        ] = self.on_scroll_event
        self.signal_handlers[
            'on_button_toggle_dropdown_scroll_event'
        ] = self.on_scroll_event
        self.signal_handlers['on_entry_text_scroll_event'] = self.on_scroll_event
        self.signal_handlers[
            'on_stored_values_popup_window_focus_out_event'
        ] = self.on_stored_values_popup_window_focus_out_event
        # For when clicking outside the popup
        self.signal_handlers[
            'on_stored_values_popup_window_button_press_event'
        ] = self.on_popup_window_button_press_event

        # Buttons for manipulating the list
        self.signal_handlers['on_button_add_clicked'] = self.on_button_add_clicked
        self.signal_handlers['on_button_edit_clicked'] = self.on_button_edit_clicked
        self.signal_handlers['on_button_remove_clicked'] = self.on_button_remove_clicked
        self.signal_handlers['on_button_up_clicked'] = self.on_button_up_clicked
        self.signal_handlers['on_button_down_clicked'] = self.on_button_down_clicked
        self.signal_handlers[
            'on_button_default_clicked'
        ] = self.on_button_default_clicked
        self.signal_handlers[
            'on_button_properties_clicked'
        ] = self.path_entry._on_button_properties_clicked

    def popup(self):
        """
        Makes the popup visible.

        """
        # Calling super popup
        PathChooserPopup.popup(self)
        self.popup_window.grab_focus()

        if not self.treeview.has_focus():
            self.treeview.grab_focus()
        if not self.popup_grab_window():
            self.popup_window.hide()
            return

        self.popup_window.grab_add()
        # Set value selected if it exists
        self.set_selected_value(
            path_without_trailing_path_sep(self.path_entry.get_text())
        )

    #################
    # Callbacks
    #################

    def on_stored_values_popup_window_focus_out_event(self, entry, event):
        """
        Popup sometimes loses the focus to the text entry, e.g. when right click
        shows a popup menu on a row. This regains the focus.
        """
        self.popup_grab_window()
        return True

    def on_scroll_event(self, widget, event):
        """
        Handles scroll events from text entry, toggle button and treeview

        """

        swap = event.get_state() & Gdk.ModifierType.CONTROL_MASK
        scroll_window = event.get_state() & Gdk.ModifierType.SHIFT_MASK
        self.handle_list_scroll(
            _next=event.direction == Gdk.ScrollDirection.DOWN,
            set_entry=widget != self.treeview,
            swap=swap,
            scroll_window=scroll_window,
        )
        return True

    def on_buttonbox_key_press_event(self, widget, event):
        """
        Handles when Escape or ALT+arrow up is pressed when focus
        is on any of the buttons in the popup
        """
        keyval = event.keyval
        state = event.get_state() & Gtk.accelerator_get_default_mod_mask()
        if keyval == Gdk.KEY_Escape or (
            key_is_up(keyval) and state == Gdk.ModifierType.MOD1_MASK
        ):
            self.popdown()
            return True
        return False

    # --------------------------------------------------
    # Funcs and callbacks on the buttons to manipulate the list
    # --------------------------------------------------
    def add_current_value_to_saved_list(self):
        text = self.path_entry.get_text()
        text = path_without_trailing_path_sep(text)
        values = self.get_values()
        if text in values:
            # Make the matching value selected
            self.set_selected_value(text)
            self.handle_list_scroll()
            return True
        self.add_values([text], scroll_to_row=True, append=False, emit_signal=True)

    def edit_selected_path(self):
        path = self.get_selection_path()
        if path:
            self.on_edit_path(path, self.tree_column)

    def on_button_add_clicked(self, widget):
        self.add_current_value_to_saved_list()
        self.popup()

    def on_button_edit_clicked(self, widget):
        self.edit_selected_path()

    def on_button_remove_clicked(self, widget):
        self.remove_selected_path()
        return True

    def on_button_up_clicked(self, widget):
        self.handle_list_scroll(_next=False, swap=True)

    def on_button_down_clicked(self, widget):
        self.handle_list_scroll(_next=True, swap=True)

    def on_button_default_clicked(self, widget):
        if self.default_text:
            self.set_text(self.default_text, trigger_event=True)


class PathCompletionPopup(CompletionList, PathChooserPopup):
    """

    The auto completion popup

    """

    def __init__(self, builder, path_entry, max_visible_rows):
        self.builder = builder
        self.treeview = self.builder.get_object('completion_treeview')
        self.popup_window = self.builder.get_object('completion_popup_window')
        self.path_entry = path_entry
        self.text_entry = path_entry.text_entry
        self.show_hidden_files = False

        self.signal_handlers = {}
        PathChooserPopup.__init__(self, 1, max_visible_rows, self.text_entry)
        CompletionList.__init__(self)

        # Add signal handlers
        self.signal_handlers[
            'on_completion_treeview_scroll_event'
        ] = self.on_scroll_event
        self.signal_handlers[
            'on_completion_popup_window_focus_out_event'
        ] = self.on_completion_popup_window_focus_out_event

        # For when clicking outside the popup
        self.signal_handlers[
            'on_completion_popup_window_button_press_event'
        ] = self.on_popup_window_button_press_event

    def popup(self):
        """
        Makes the popup visible.

        """
        PathChooserPopup.popup(self)
        self.popup_window.grab_focus()

        if not self.treeview.has_focus():
            self.treeview.grab_focus()

        if not self.popup_grab_window():
            self.popup_window.hide()
            return

        self.popup_window.grab_add()
        self.text_entry.grab_focus()
        self.text_entry.set_position(len(self.path_entry.text_entry.get_text()))
        self.set_selected_value(
            path_without_trailing_path_sep(self.path_entry.get_text()),
            select_first=True,
        )

    #################
    # Callbacks
    #################

    def on_completion_popup_window_focus_out_event(self, entry, event):
        """
        Popup sometimes loses the focus to the text entry, e.g. when right click
        shows a popup menu on a row. This regains the focus.
        """
        self.popup_grab_window()
        return True

    def on_scroll_event(self, widget, event):
        """
        Handles scroll events from the treeview

        """
        x, y = event.window.get_pointer()
        self.handle_list_scroll(
            _next=event.direction == Gdk.ScrollDirection.DOWN,
            set_entry=widget != self.treeview,
            scroll_window=True,
        )
        path = self.treeview.get_path_at_pos(int(x), int(y))
        if path:
            self.handle_list_scroll(path=path[0], _next=None)
        return True


class PathAutoCompleter:
    def __init__(self, builder, path_entry, max_visible_rows):
        self.completion_popup = PathCompletionPopup(
            builder, path_entry, max_visible_rows
        )
        self.path_entry = path_entry
        self.dirs_cache = {}
        self.use_popup = False
        self.auto_complete_enabled = True
        self.signal_handlers = self.completion_popup.signal_handlers

        self.signal_handlers[
            'on_completion_popup_window_key_press_event'
        ] = self.on_completion_popup_window_key_press_event
        self.signal_handlers[
            'on_entry_text_delete_text'
        ] = self.on_entry_text_delete_text
        self.signal_handlers[
            'on_entry_text_insert_text'
        ] = self.on_entry_text_insert_text
        self.accelerator_string = Gtk.accelerator_name(Gdk.KEY_Tab, 0)

    def on_entry_text_insert_text(self, entry, new_text, new_text_length, position):
        if self.path_entry.get_realized():
            cur_text = self.path_entry.get_text()
            pos = entry.get_position()
            new_complete_text = cur_text[:pos] + new_text + cur_text[pos:]
            # Remove all values from the list that do not start with new_complete_text
            self.completion_popup.reduce_values(new_complete_text)
            self.completion_popup.set_selected_value(
                new_complete_text, select_first=True
            )
            if self.completion_popup.is_popped_up():
                self.completion_popup.set_window_position_and_size()

    def on_entry_text_delete_text(self, entry, start, end):
        """
        Do completion when characters are removed

        """
        if self.completion_popup.is_popped_up():
            cur_text = self.path_entry.get_text()
            new_complete_text = cur_text[:start] + cur_text[end:]
            self.do_completion(value=new_complete_text, forward_completion=False)

    def set_use_popup(self, use):
        self.use_popup = use

    def on_completion_popup_window_key_press_event(self, entry, event):
        """
        Handles key pressed events on the auto-completion popup window
        """
        # If on_completion_treeview_key_press_event handles the event, do nothing
        ret = self.completion_popup.on_completion_treeview_key_press_event(entry, event)
        if ret:
            return ret
        keyval = event.keyval
        state = event.get_state() & Gtk.accelerator_get_default_mod_mask()
        if (
            self.is_auto_completion_accelerator(keyval, state)
            and self.auto_complete_enabled
        ):
            values_count = self.completion_popup.get_values_count()
            if values_count == 1:
                self.do_completion()
            else:
                self.completion_popup.handle_list_scroll(_next=True)
            return True
        # Buggy stuff (in pygobject?) causing type mismatch between EventKey and GdkEvent. Convert manually...
        n = Gdk.Event()
        n.type = event.type
        n.window = event.window
        n.send_event = event.send_event
        n.time = event.time
        n.state = event.state
        n.keyval = event.keyval
        n.length = event.length
        n.string = event.string
        n.hardware_keycode = event.hardware_keycode
        n.group = event.group
        n.is_modifier = event.is_modifier
        self.path_entry.text_entry.emit('key-press-event', n)

    def is_auto_completion_accelerator(self, keyval, state):
        return Gtk.accelerator_name(keyval, state) == self.accelerator_string

    def do_completion(self, value=None, forward_completion=True):
        if not value:
            value = self.path_entry.get_text()
        self.path_entry.text_entry.set_position(len(value))
        opts = {}
        opts['show_hidden_files'] = self.completion_popup.show_hidden_files
        opts['completion_text'] = value
        opts['forward_completion'] = forward_completion
        self._start_completion(opts)

    def _start_completion(self, args):
        args = get_completion_paths(args)
        self._end_completion(args)

    def _end_completion(self, args):
        value = args['completion_text']
        paths = args['paths']

        if args['forward_completion']:
            common_prefix = os.path.commonprefix(paths)
            if len(common_prefix) > len(value):
                self.path_entry.set_text(
                    common_prefix, set_file_chooser_folder=True, trigger_event=True
                )

        self.path_entry.text_entry.set_position(len(self.path_entry.get_text()))
        self.completion_popup.set_values(paths, preserve_selection=True)

        if self.use_popup and len(paths) > 1:
            self.completion_popup.popup()
        elif self.completion_popup.is_popped_up() and args['forward_completion']:
            self.completion_popup.popdown()


class PathChooserComboBox(Gtk.Box, StoredValuesPopup, GObject.GObject):

    __gsignals__ = {
        signal: (SignalFlags.RUN_FIRST, GObject.TYPE_NONE, (object,))
        for signal in [
            'text-changed',
            'accelerator-set',
            'max-rows-changed',
            'list-value-added',
            'list-value-removed',
            'list-values-changed',
            'list-values-reordered',
            'show-path-entry-toggled',
            'show-filechooser-toggled',
            'show-hidden-files-toggled',
            'show-folder-name-on-button',
            'auto-complete-enabled-toggled',
        ]
    }

    def __init__(
        self,
        max_visible_rows=20,
        auto_complete=True,
        use_completer_popup=True,
        parent=None,
    ):
        Gtk.Box.__init__(self)
        GObject.GObject.__init__(self)
        self.list_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=0)
        self._stored_values_popping_down = False
        self.filechooser_visible = True
        self.filechooser_enabled = True
        self.path_entry_visible = True
        self.properties_enabled = True
        self.show_folder_name_on_button = False
        self.setting_accelerator_key = False
        self.builder = Gtk.Builder()
        self.parent = parent
        self.popup_buttonbox = self.builder.get_object('buttonbox')
        self.builder.add_from_file(
            resource_filename(
                __package__, os.path.join('glade', 'path_combo_chooser.ui')
            )
        )
        self.button_toggle = self.builder.get_object('button_toggle_dropdown')
        self.text_entry = self.builder.get_object('entry_text')
        self.open_filechooser_dialog_button = self.builder.get_object(
            'button_open_dialog'
        )
        self.filechooser_button = self.open_filechooser_dialog_button
        self.filechooserdialog = self.builder.get_object('filechooserdialog')
        self.filechooserdialog.set_transient_for(self.parent)
        self.filechooser_widget = self.builder.get_object('filechooser_widget')
        self.folder_name_label = self.builder.get_object('folder_name_label')
        self.default_text = None
        self.button_properties = self.builder.get_object('button_properties')

        self.combobox_window = self.builder.get_object('combobox_window')
        self.combo_hbox = self.builder.get_object('entry_combobox_hbox')
        # Change the parent of the hbox from the glade Window to this hbox.
        self.combobox_window.remove(self.combo_hbox)
        self.combobox_window = self.get_window()
        self.add(self.combo_hbox)
        StoredValuesPopup.__init__(
            self, self.builder, self, max_visible_rows, self.combo_hbox
        )

        self.tooltips = Gtk.Tooltip()
        self.auto_completer = PathAutoCompleter(self.builder, self, max_visible_rows)
        self.auto_completer.set_use_popup(use_completer_popup)
        self.auto_completer.auto_complete_enabled = auto_complete
        self._setup_config_dialog()

        signal_handlers = {
            'on_button_toggle_dropdown_toggled': self._on_button_toggle_dropdown_toggled,
            'on_entry_text_key_press_event': self._on_entry_text_key_press_event,
            'on_stored_values_popup_window_hide': self._on_stored_values_popup_window_hide,
            'on_button_toggle_dropdown_button_press_event': self._on_button_toggle_dropdown_button_press_event,
            'on_entry_combobox_hbox_realize': self._on_entry_combobox_hbox_realize,
            'on_button_open_dialog_clicked': self._on_button_open_dialog_clicked,
            'on_entry_text_focus_out_event': self._on_entry_text_focus_out_event,
            'on_entry_text_changed': self.on_entry_text_changed,
        }
        signal_handlers.update(self.signal_handlers)
        signal_handlers.update(self.auto_completer.signal_handlers)
        signal_handlers.update(self.config_dialog_signal_handlers)
        self.builder.connect_signals(signal_handlers)

    def get_text(self):
        """
        Get the current text in the Entry
        """
        return self.text_entry.get_text()

    def set_text(
        self,
        text,
        set_file_chooser_folder=True,
        cursor_end=True,
        default_text=False,
        trigger_event=False,
    ):
        """
        Set the text for the entry.

        """
        old_text = self.text_entry.get_text()
        # We must block the "delete-text" signal to avoid the signal handler being called
        self.text_entry.handler_block_by_func(
            self.auto_completer.on_entry_text_delete_text
        )
        self.text_entry.set_text(text)
        self.text_entry.handler_unblock_by_func(
            self.auto_completer.on_entry_text_delete_text
        )

        self.text_entry.select_region(0, 0)
        self.text_entry.set_position(len(text) if cursor_end else 0)
        self.set_selected_value(text, select_first=True)
        self.combo_hbox.set_tooltip_text(text)
        if default_text:
            self.default_text = text
            self.button_default.set_tooltip_text(
                'Restore the default value in the text entry:\n%s' % self.default_text
            )
            self.button_default.set_sensitive(True)
        # Set text for the filechooser dialog button
        folder_name = ''
        if self.show_folder_name_on_button or not self.path_entry_visible:
            folder_name = path_without_trailing_path_sep(text)
            if folder_name != '/' and os.path.basename(folder_name):
                folder_name = os.path.basename(folder_name)
        self.folder_name_label.set_text(folder_name)
        # Only trigger event if text has changed
        if old_text != text and trigger_event:
            self.on_entry_text_changed(self.text_entry)

    def set_sensitive(self, sensitive):
        """
        Set the path chooser widgets sensitive

        :param sensitive: if the widget should be sensitive
        :type  sensitive: bool

        """
        self.text_entry.set_sensitive(sensitive)
        self.filechooser_button.set_sensitive(sensitive)
        self.button_toggle.set_sensitive(sensitive)

    def get_accelerator_string(self):
        return self.auto_completer.accelerator_string

    def set_accelerator_string(self, accelerator):
        """
        Set the accelerator string to trigger auto-completion
        """
        if accelerator is None:
            return
        try:
            # Verify that the accelerator can be parsed
            keyval, mask = Gtk.accelerator_parse(self.auto_completer.accelerator_string)
            self.auto_completer.accelerator_string = accelerator
        except TypeError as ex:
            raise TypeError('TypeError when setting accelerator string: %s' % ex)

    def get_auto_complete_enabled(self):
        return self.auto_completer.auto_complete_enabled

    def set_auto_complete_enabled(self, enable):
        if not isinstance(enable, bool):
            return
        self.auto_completer.auto_complete_enabled = enable

    def get_show_folder_name_on_button(self):
        return self.show_folder_name_on_button

    def set_show_folder_name_on_button(self, show):
        if not isinstance(show, bool):
            return
        self.show_folder_name_on_button = show
        self._set_path_entry_filechooser_widths()

    def get_filechooser_button_enabled(self):
        return self.filechooser_enabled

    def set_filechooser_button_enabled(self, enable):
        """
        Enable/disable the filechooser button.

        By setting filechooser disabled, in will not be possible
        to change the settings in the properties.
        """
        if not isinstance(enable, bool):
            return
        self.filechooser_enabled = enable
        if not enable:
            self.set_filechooser_button_visible(False, update=False)

    def get_filechooser_button_visible(self):
        return self.filechooser_visible

    def set_filechooser_button_visible(self, visible, update=True):
        """
        Set file chooser button entry visible
        """
        if not isinstance(visible, bool):
            return
        if update:
            self.filechooser_visible = visible
        if visible and not self.filechooser_enabled:
            return
        if visible:
            self.filechooser_button.show()
        else:
            self.filechooser_button.hide()
        # Update width properties
        self._set_path_entry_filechooser_widths()

    def get_path_entry_visible(self):
        return self.path_entry_visible

    def set_path_entry_visible(self, visible):
        """
        Set the path entry visible
        """
        if not isinstance(visible, bool):
            return
        self.path_entry_visible = visible
        if visible:
            self.text_entry.show()
        else:
            self.text_entry.hide()
        self._set_path_entry_filechooser_widths()

    def get_show_hidden_files(self):
        return self.auto_completer.completion_popup.show_hidden_files

    def set_show_hidden_files(self, show, do_completion=False, emit_event=False):
        """
        Enable/disable showing hidden files on path completion
        """
        if not isinstance(show, bool):
            return
        self.auto_completer.completion_popup.show_hidden_files = show
        if do_completion:
            self.auto_completer.do_completion()
        if emit_event:
            self.emit('show-hidden-files-toggled', show)

    def set_enable_properties(self, enable):
        """
        Enable/disable the config properties
        """
        if not isinstance(enable, bool):
            return
        self.properties_enabled = enable
        if self.properties_enabled:
            self.popup_buttonbox.add(self.button_properties)
        else:
            self.popup_buttonbox.remove(self.button_properties)

    def set_auto_completer_func(self, func):
        """
        Set the function to be called when the auto completion
        accelerator is triggered.
        """
        self.auto_completer._start_completion = func

    def complete(self, args):
        """
        Perform the auto completion with the provided paths
        """
        self.auto_completer._end_completion(args)

    ##############
    # Callbacks and internal functions
    ##############

    def on_entry_text_changed(self, entry):
        self.emit('text-changed', self.get_text())

    def _on_entry_text_focus_out_event(self, widget, event):
        # FIXME: This causes text to be deselected on right-click.
        # Update text on the button label
        self.set_text(self.get_text())

    def _set_path_entry_filechooser_widths(self):
        if self.path_entry_visible:
            self.combo_hbox.set_child_packing(
                self.filechooser_button, 0, 0, 0, Gtk.PackType.START
            )
            width, height = self.folder_name_label.get_size_request()
            width = 120
            if not self.show_folder_name_on_button:
                width = 0
            self.folder_name_label.set_size_request(width, height)
            self.combo_hbox.set_child_packing(
                self.filechooser_button, 0, 0, 0, Gtk.PackType.START
            )
        else:
            self.combo_hbox.set_child_packing(
                self.filechooser_button, 1, 1, 0, Gtk.PackType.START
            )
            self.folder_name_label.set_size_request(-1, -1)
        # Update text on the button label
        self.set_text(self.get_text())

    def _on_entry_combobox_hbox_realize(self, widget):
        """Must do this when the widget is realized"""
        self.set_filechooser_button_visible(self.filechooser_visible)
        self.set_path_entry_visible(self.path_entry_visible)

    def _on_button_open_dialog_clicked(self, widget):
        self.filechooser_widget.set_current_folder(self.get_text())
        response_id = self.filechooserdialog.run()

        if response_id == 0:
            text = self.filechooser_widget.get_filename()
            self.set_text(text, trigger_event=True)
        self.filechooserdialog.hide()

    def _on_entry_text_key_press_event(self, widget, event):
        """
        Listen to key events on the entry widget.

        Arrow up/down will change the value of the entry according to the
        current selection in the list.
        Enter will show the popup.

        Return True whenever we want no other event listeners to be called.

        """
        # on_entry_text_key_press_event Errors follow here when pressing ALT key while popup is visible")
        keyval = event.keyval
        state = event.get_state() & Gtk.accelerator_get_default_mod_mask()
        ctrl = event.get_state() & Gdk.ModifierType.CONTROL_MASK

        # Select new row with arrow up/down is pressed
        if key_is_up_or_down(keyval):
            self.handle_list_scroll(_next=key_is_down(keyval), set_entry=True)
            return True
        elif self.auto_completer.is_auto_completion_accelerator(keyval, state):
            if self.auto_completer.auto_complete_enabled:
                self.auto_completer.do_completion()
                return True
        # Show popup when Enter is pressed
        elif key_is_enter(keyval):
            # This sets the toggle active which results in
            # on_button_toggle_dropdown_toggled being called which initiates the popup
            self.button_toggle.set_active(True)
            return True
        elif ctrl:
            # Swap the show hidden files value on CTRL-h
            if is_ascii_value(keyval, 'h'):
                # Set show/hide hidden files
                self.set_show_hidden_files(
                    not self.get_show_hidden_files(), emit_event=True
                )
                return True
            elif is_ascii_value(keyval, 's'):
                super().add_current_value_to_saved_list()
                return True
            elif is_ascii_value(keyval, 'd'):
                # Set the default value in the text entry
                self.set_text(self.default_text, trigger_event=True)
                return True
        return False

    def _on_button_toggle_dropdown_toggled(self, button):
        """
        Shows the popup when clicking the toggle button.
        """
        if self._stored_values_popping_down:
            return
        self.popup()

    def _on_stored_values_popup_window_hide(self, popup):
        """Make sure the button toggle is removed when popup is closed"""
        self._stored_values_popping_down = True
        self.button_toggle.set_active(False)
        self._stored_values_popping_down = False

    ##############
    # Config dialog
    ##############

    def _on_button_toggle_dropdown_button_press_event(self, widget, event):
        """Show config when right clicking dropdown toggle button"""
        if not self.properties_enabled:
            return False
        # This is right click
        if event.button == 3:
            self._on_button_properties_clicked(widget)
            return True

    def _on_button_properties_clicked(self, widget):
        self.popdown()
        self.enable_completion.set_active(self.get_auto_complete_enabled())
        # Set the value of the label to the current accelerator
        keyval, mask = Gtk.accelerator_parse(self.auto_completer.accelerator_string)
        self.accelerator_label.set_text(Gtk.accelerator_get_label(keyval, mask))
        self.visible_rows.set_value(self.get_max_popup_rows())
        self.show_filechooser_checkbutton.set_active(
            self.get_filechooser_button_visible()
        )
        self.show_path_entry_checkbutton.set_active(self.path_entry_visible)
        self.show_hidden_files_checkbutton.set_active(self.get_show_hidden_files())
        self.show_folder_name_on_button_checkbutton.set_active(
            self.get_show_folder_name_on_button()
        )
        self._set_properties_widgets_sensitive(True)
        self.config_dialog.show_all()

    def _set_properties_widgets_sensitive(self, val):
        self.enable_completion.set_sensitive(val)
        self.config_short_cuts_frame.set_sensitive(val)
        self.config_general_frame.set_sensitive(val)
        self.show_hidden_files_checkbutton.set_sensitive(val)

    def _setup_config_dialog(self):
        self.config_dialog = self.builder.get_object('completion_config_dialog')
        self.enable_completion = self.builder.get_object(
            'enable_auto_completion_checkbutton'
        )
        self.show_filechooser_checkbutton = self.builder.get_object(
            'show_filechooser_checkbutton'
        )
        self.show_path_entry_checkbutton = self.builder.get_object(
            'show_path_entry_checkbutton'
        )
        set_key_button = self.builder.get_object('set_completion_accelerator_button')
        default_set_accelerator_tooltip = set_key_button.get_tooltip_text()
        self.config_short_cuts_frame = self.builder.get_object(
            'config_short_cuts_frame'
        )
        self.config_general_frame = self.builder.get_object('config_general_frame')
        self.accelerator_label = self.builder.get_object('completion_accelerator_label')
        self.visible_rows = self.builder.get_object('visible_rows_spinbutton')
        self.visible_rows_label = self.builder.get_object('visible_rows_label')
        self.show_hidden_files_checkbutton = self.builder.get_object(
            'show_hidden_files_checkbutton'
        )
        self.show_folder_name_on_button_checkbutton = self.builder.get_object(
            'show_folder_name_on_button_checkbutton'
        )
        self.config_dialog.set_transient_for(self.parent)

        def on_close(widget, event=None):
            if not self.setting_accelerator_key:
                self.config_dialog.hide()
            else:
                stop_setting_accelerator()
            return True

        def on_enable_completion_toggled(widget):
            self.set_auto_complete_enabled(self.enable_completion.get_active())
            self.emit(
                'auto-complete-enabled-toggled', self.enable_completion.get_active()
            )

        def on_show_filechooser_toggled(widget):
            self.set_filechooser_button_visible(
                self.show_filechooser_checkbutton.get_active()
            )
            self.emit(
                'show-filechooser-toggled',
                self.show_filechooser_checkbutton.get_active(),
            )
            self.show_folder_name_on_button_checkbutton.set_sensitive(
                self.show_path_entry_checkbutton.get_active()
                and self.show_filechooser_checkbutton.get_active()
            )
            if not self.filechooser_visible and not self.path_entry_visible:
                self.show_path_entry_checkbutton.set_active(True)
                on_show_path_entry_toggled(None)

        def on_show_path_entry_toggled(widget):
            self.set_path_entry_visible(self.show_path_entry_checkbutton.get_active())
            self.emit(
                'show-path-entry-toggled', self.show_path_entry_checkbutton.get_active()
            )
            self.show_folder_name_on_button_checkbutton.set_sensitive(
                self.show_path_entry_checkbutton.get_active()
                and self.show_filechooser_checkbutton.get_active()
            )
            if not self.filechooser_visible and not self.path_entry_visible:
                self.show_filechooser_checkbutton.set_active(True)
                on_show_filechooser_toggled(None)

        def on_show_folder_name_on_button(widget):
            self.set_show_folder_name_on_button(
                self.show_folder_name_on_button_checkbutton.get_active()
            )
            self._set_path_entry_filechooser_widths()
            self.emit(
                'show-folder-name-on-button',
                self.show_folder_name_on_button_checkbutton.get_active(),
            )

        def on_show_hidden_files_toggled(widget):
            self.set_show_hidden_files(
                self.show_hidden_files_checkbutton.get_active(), emit_event=True
            )

        def on_max_rows_changed(widget):
            self.set_max_popup_rows(self.visible_rows.get_value_as_int())
            self.emit('max-rows-changed', self.visible_rows.get_value_as_int())

        def set_accelerator(widget):
            self.setting_accelerator_key = True
            set_key_button.set_tooltip_text(
                'Press the accelerator keys for triggering auto-completion'
            )
            self._set_properties_widgets_sensitive(False)
            return True

        def stop_setting_accelerator():
            self.setting_accelerator_key = False
            self._set_properties_widgets_sensitive(True)
            set_key_button.set_active(False)
            # Restore default tooltip
            set_key_button.set_tooltip_text(default_set_accelerator_tooltip)

        def on_completion_config_dialog_key_release_event(widget, event):
            # We are listening for a new key
            if set_key_button.get_active():
                state = event.get_state() & Gtk.accelerator_get_default_mod_mask()
                accelerator_mask = state.numerator
                # If e.g. only CTRL key is pressed.
                if not Gtk.accelerator_valid(event.keyval, accelerator_mask):
                    accelerator_mask = 0
                self.auto_completer.accelerator_string = Gtk.accelerator_name(
                    event.keyval, accelerator_mask
                )
                self.accelerator_label.set_text(
                    Gtk.accelerator_get_label(event.keyval, accelerator_mask)
                )
                self.emit('accelerator-set', self.auto_completer.accelerator_string)
                stop_setting_accelerator()
                return True
            else:
                keyval = event.keyval
                ctrl = event.get_state() & Gdk.ModifierType.CONTROL_MASK
                if ctrl:
                    # Set show/hide hidden files
                    if is_ascii_value(keyval, 'h'):
                        self.show_hidden_files_checkbutton.set_active(
                            not self.get_show_hidden_files()
                        )
                        return True

        def on_set_completion_accelerator_button_clicked(widget):
            if not set_key_button.get_active():
                stop_setting_accelerator()
                return True

        self.config_dialog_signal_handlers = {
            'on_enable_auto_completion_checkbutton_toggled': on_enable_completion_toggled,
            'on_show_filechooser_checkbutton_toggled': on_show_filechooser_toggled,
            'on_show_path_entry_checkbutton_toggled': on_show_path_entry_toggled,
            'on_show_folder_name_on_button_checkbutton_toggled': on_show_folder_name_on_button,
            'on_config_dialog_button_close_clicked': on_close,
            'on_visible_rows_spinbutton_value_changed': on_max_rows_changed,
            'on_completion_config_dialog_delete_event': on_close,
            'on_set_completion_accelerator_button_pressed': set_accelerator,
            'on_completion_config_dialog_key_release_event': on_completion_config_dialog_key_release_event,
            'on_set_completion_accelerator_button_clicked': on_set_completion_accelerator_button_clicked,
            'on_show_hidden_files_checkbutton_toggled': on_show_hidden_files_toggled,
        }


GObject.type_register(PathChooserComboBox)


if __name__ == '__main__':
    import signal

    # necessary to exit with CTRL-C (https://bugzilla.gnome.org/show_bug.cgi?id=622084)
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    import sys

    w = Gtk.Window()
    w.set_position(Gtk.WindowPosition.CENTER)
    w.set_size_request(600, -1)
    w.set_title('ComboEntry example')
    w.connect('delete-event', Gtk.main_quit)

    box1 = Gtk.Box.new(Gtk.Orientation.VERTICAL, spacing=0)

    def get_resource2(filename):
        return f'{os.path.abspath(os.path.dirname(sys.argv[0]))}/glade/{filename}'

    # Override get_resource which fetches from deluge install
    # get_resource = get_resource2

    entry1 = PathChooserComboBox(max_visible_rows=15)
    entry2 = PathChooserComboBox()

    box1.add(entry1)
    box1.add(entry2)

    test_paths = [
        '/home/bro/Downloads',
        '/media/Movies-HD',
        '/media/torrent/in',
        '/media/Live-show/Misc',
        '/media/Live-show/Consert',
        '/media/Series/1/',
        '/media/Series/2',
        '/media/Series/17',
        '/media/Series/18',
        '/media/Series/19',
    ]

    entry1.add_values(test_paths)
    entry1.set_text('/home/bro/', default_text=True)
    entry2.set_text(
        '/home/bro/programmer/deluge/deluge-yarss-plugin/build/lib/yarss2/include/bs4/tests/',
        cursor_end=False,
    )

    entry2.set_filechooser_button_visible(False)
    # entry2.set_enable_properties(False)
    entry2.set_filechooser_button_enabled(False)

    def list_value_added_event(widget, values):
        print('Current list values:', widget.get_values())

    entry1.connect('list-value-added', list_value_added_event)
    entry2.connect('list-value-added', list_value_added_event)
    w.add(box1)
    w.show_all()
    Gtk.main()



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from deluge.ui.console.modes.basemode import add_string, mkpad, mkpanel
from deluge.ui.console.utils.colors import get_color_pair

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)


class BaseWindow:
    """
    BaseWindow creates a curses screen to be used for showing panels and popup dialogs
    """

    def __init__(self, title, width, height, posy=0, posx=0, encoding=None):
        """
        Args:
            title (str): The title of the panel
            width (int): Width of the panel
            height (int): Height of the panel
            posy (int): Position of the panel's first row relative to the terminal screen
            posx (int): Position of the panel's first column relative to the terminal screen
            encoding (str): Terminal encoding
        """
        self.title = title
        self.posy, self.posx = posy, posx
        if encoding is None:
            from deluge import component

            encoding = component.get('ConsoleUI').encoding
        self.encoding = encoding

        self.panel = mkpanel(curses.COLOR_GREEN, height, width, posy, posx)
        self.outer_screen = self.panel.window()
        self.outer_screen.bkgdset(0, curses.COLOR_RED)
        by, bx = self.outer_screen.getbegyx()
        self.screen = mkpad(get_color_pair('white', 'black'), height - 1, width - 2)
        self._height, self._width = self.outer_screen.getmaxyx()

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    def add_string(self, row, string, scr=None, **kwargs):
        scr = scr if scr else self.screen
        add_string(row, string, scr, self.encoding, **kwargs)

    def hide(self):
        self.panel.hide()

    def show(self):
        self.panel.show()

    def hidden(self):
        return self.panel.hidden()

    def set_title(self, title):
        self.title = title

    @property
    def visible_content_pane_size(self):
        y, x = self.outer_screen.getmaxyx()
        return (y - 2, x - 2)

    @property
    def visible_content_pane_height(self):
        y, x = self.visible_content_pane_size
        return y

    @property
    def visible_content_pane_width(self):
        y, x = self.visible_content_pane_size
        return x

    def getmaxyx(self):
        return self.screen.getmaxyx()

    def resize_window(self, rows, cols):
        self.outer_screen.resize(rows, cols)
        self.screen.resize(rows - 2, cols - 2)
        self._height, self._width = rows, cols

    def move_window(self, posy, posx):
        posy = int(posy)
        posx = int(posx)
        self.outer_screen.mvwin(posy, posx)
        self.posy = posy
        self.posx = posx
        self._height, self._width = self.screen.getmaxyx()

    def ensure_content_pane_height(self, height):
        max_y, max_x = self.screen.getmaxyx()
        if max_y < height:
            self.screen.resize(height, max_x)

    def draw_scroll_indicator(self, screen):
        content_height = self.get_content_height()
        if content_height <= self.visible_content_pane_height:
            return

        percent_scroll = float(self.lineoff) / (
            content_height - self.visible_content_pane_height
        )
        indicator_row = int(self.visible_content_pane_height * percent_scroll) + 1

        # Never greater than height
        indicator_row = min(indicator_row, self.visible_content_pane_height)
        indicator_col = self.width + 1

        add_string(
            indicator_row,
            '{!red,black,bold!}#',
            screen,
            self.encoding,
            col=indicator_col,
            pad=False,
            trim=False,
        )

    def refresh(self):
        height, width = self.visible_content_pane_size
        self.outer_screen.erase()
        self.outer_screen.border(0, 0, 0, 0)

        if self.title:
            toff = max(1, (self.width // 2) - (len(self.title) // 2))
            self.add_string(
                0,
                '{!white,black,bold!}%s' % self.title,
                scr=self.outer_screen,
                col=toff,
                pad=False,
            )

        self.draw_scroll_indicator(self.outer_screen)
        self.outer_screen.noutrefresh()

        try:
            # pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol
            # the p arguments refer to the upper left corner of the pad region to be displayed and
            # the s arguments define a clipping box on the screen within which the pad region is to be displayed.
            pminrow = self.lineoff
            pmincol = 0
            sminrow = self.posy + 1
            smincol = self.posx + 1
            smaxrow = height + self.posy
            smaxcol = width + self.posx
            self.screen.noutrefresh(
                pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol
            )
        except curses.error as ex:
            import traceback

            log.warning(
                'Error on screen.noutrefresh(%s, %s, %s, %s, %s, %s) Error: %s\nStack: %s',
                pminrow,
                pmincol,
                sminrow,
                smincol,
                smaxrow,
                smaxcol,
                ex,
                ''.join(traceback.format_stack()),
            )
